/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBMATRIX_SEQ.c                                             */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>

#include "hips_define.h"
#include "block.h"

void DBMATRIX_Precond(PhidalMatrix *A, DBPrec *P, SymbolMatrix* symbmtx, PhidalHID *BL, PhidalOptions *option)
{
#ifdef DEBUG_M
  assert(A->symmetric == option->symmetric);
#endif
  
  /** Correct the options **/
  PhidalOptions_Fix(option, BL);

  /** Print the options **/
  if(option->verbose > 0)
    PhidalOptions_Print(stdout, option);

  if(A->symmetric == 0)
    DBMATRIX_MLILUPrec(A, P, symbmtx, BL, option);
  else
    DBMATRIX_MLICCPrec(A, P, symbmtx, BL, option);
}

/* TODO : nom de la fonction : c'est aussi le nom d'un fichier */
INTS DBMATRIX_Solve(PhidalMatrix *A, DBPrec *P, PhidalHID *BL, PhidalOptions *option, COEF *rhs, COEF *x, dim_t *itertab, REAL *resnormtab) {
  chrono_t t1, t2;
  t1  = dwalltime();

  /*option->krylov_method = 0;*/ /* !! */

  /** The first level is always fgmres **/
  if(option->krylov_method != 1 || (option->forwardlev > 0 && option->schur_method != 0))
    {
      CHECK_RETURN(HIPS_Fgmresd_PH_DB(option->verbose, option->tol, option->itmax, A, P, BL, option, rhs, x, option->fd, itertab, resnormtab));
    }
  else
    {
      CHECK_RETURN(HIPS_PCG_PH_DB(option->verbose, option->tol, option->itmax, A, P, BL, option, rhs, x, option->fd, itertab, resnormtab));
    }

  t2  = dwalltime();

  /* if(option->verbose > 0) */
  fprintfv(1, stdout, "Number of outer iterations %d \n", itertab[0]);
  
  fprintfv(1, stdout, "\n SOLVED IN %g \n", t2-t1);
  return HIPS_SUCCESS;
}

