/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBMatrix_Build.c                                           */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h> /* strcmp */

#include "block.h"

#ifdef PARALLEL
#include "db_parallel.h"
#endif

void CMD(_UDBDistrMatrix,Build)(PhidalMatrix* mat, 
				dim_t tli, dim_t tlj, dim_t bri, dim_t brj, 
				char *UPLO_L, char *UPLO_U, 
				int_t locally_nbr, 
				PhidalMatrix* PhidalL, PhidalMatrix* PhidalU, 
				_UDBDistrMatrix* DBM, 
				SymbolMatrix* symbmtx, _PhidalDistrHID *_DBL,
				int_t levelnum, int_t levelnum_l,
				alloc_t alloc, flag_t filling) {

  DBM->L = (_DBDistrMatrix*)malloc(sizeof(_DBDistrMatrix));
  S(DBM->L)->symmetric = mat->symmetric;

  CMD(_DBDistrMatrix,Setup)(tli, tlj, bri, brj, UPLO_L, "N", locally_nbr, 
			   DBM->L, symbmtx, _DBL, levelnum, levelnum_l, alloc);
  
  CMD(_DBDistrMatrix,Copy)(mat, _DBL, PhidalL, DBM->L, UPLO_L, levelnum, levelnum_l, filling);


  if (mat->symmetric == 1) {

    DBM->U = NULL;

  } else {

    DBM->U = (_DBDistrMatrix*)malloc(sizeof(_DBDistrMatrix));
    S(DBM->U)->symmetric = mat->symmetric;
    
    CMD(_DBDistrMatrix,SetupV)(DBM->L, DBM->U);
    CMD(_DBDistrMatrix,Transpose)(DBM->U);
    
    CMD(_DBDistrMatrix,Copy)(mat, _DBL, PhidalU, DBM->U, UPLO_U, levelnum, levelnum_l, filling);

  } 

}
