/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBMatrix_ICCT.c                                            */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_parallel.h"
#include "db_parallel.h"
#include "block.h"

#define Dprintf(arg...)

void DBMatrix_ICCT_Restrict(int START, DBMatrix *L, PhidalHID *BL, PhidalOptions *option);

void DBMatrix_ICCT(DBMatrix *L, PhidalHID *BL, PhidalOptions *option)
{
  /******************************************************************************************/
  /* This function computes the incomplete factorization A = L.D.Lt of a symmetric matrix   */
  /* On entry:                                                                              */
  /*   L is the lower triangular part of the matrix to factorize in CSC format              */
  /*   job == 0 : the shared block are freed on the non-leader                              */
  /*   processors otherwise they stay in memory (need this when to use                      */
  /* L^-1.M in multilevel recursion)                                                        */
  /* on return:                                                                             */
  /* L the strictly lower triangular factor (unitary on the diagonal not stored) in CSC     */
  /* D is the INVERSE of the diagonal factor                                                */
  /* NOTE: L is sorted by row index on return                                               */
  /******************************************************************************************/
#ifdef DEBUG_M
  assert(S(L)->symmetric == 1);
#endif
  DBMatrix_ICCT_Restrict(L->tli, L, BL, option);
}



void DBMatrix_ICCT_Restrict(int START, DBMatrix *L, PhidalHID *BL, PhidalOptions *option)
{

  dim_t i,j, k;
  int nk;
  dim_t *jak; int *rindk;
  VSolverMatrix **rak;
  int ii, jj, kk;
  VSolverMatrix *csL;
  VSolverMatrix **csrtab1, **list1;
  VSolverMatrix **csrtab2, **list2;
  int *listindex;
  int nnb;
  CellDB *firstcol;
  DBMatrix *LL=L;

  COEF *E, *W;
  int *tabA, *tabC;
 
  printf("DBMatrix_ICCT\n");

  E = (COEF *)malloc(sizeof(COEF)*LL->coefmax);
  assert(E != NULL);
 
  W = (COEF *)malloc(sizeof(COEF)*LL->coefmax); /* TODO */
  assert(W != NULL);

  tabA = (int*)malloc(sizeof(int)*LL->dim1);
  tabC = (int*)malloc(sizeof(int)*LL->dim1);












  /* repère A */




  for(k=START;k<=LL->brj;k++)
    {
      Dprintf("k=%d\n", k);
      /*** Factorize the column block k ***/
	
      csL =  LL->ca[ LL->cia[k] ];

      if (k>L->tlj) {
	Dprintf(" DBICCprod\n", k, LL->brj);
	
	DB_ICCprod(k, -1,
		   LL,
		   nk, jak, rak,
		   tabA, tabC,
		   E, NULL, W);
	
      }
      /* Repère B */









     
      /*** Factorize the diagonal block matrix of the column k ***/
      /* kk = BL->block_index[k]-BL->block_index[LL->tlj]; unused */

      Dprintf("Facto du id=%d\n", LL->ca[LL->cia[k]] - LL->ca[0]);
      VS_ICCT(csL, /* F */E, W); /* todo : simplifier, ce n'est qu'un bloc dense dans S */

    
      /*** Divide the column block matrices by the lower triangular
	   block diagonal factor ***/
      for(ii=LL->cia[k]+1;ii<LL->cia[k+1];ii++) {
	i = LL->cja[ii];
	
	Dprintf("Div de id=%d par id=%d\n", LL->ca[ii] - LL->ca[0], LL->ca[LL->cia[k]] - LL->ca[0]);

	
	VS_InvLT(1, csL, LL->ca[ii], W); /* todo : simplifier, ce n'est qu'un bloc dense dans S */
      }





      
    } /* (k=START;k<=LL->brj;k++) */
  
  free(E);
  free(W);
  free(tabA);
  free(tabC);

  /* /\* if (k==START+1) *\/ { printDBMatrix(L); exit(1); } */
}

