/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBPrec_pic.c                                               */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>

#include "block.h"

#define need(arg) if ((arg) == -1) { /* printfv(5, "no %s\n", #arg); */ return -1; }

/* Preconditionner size (in nnz) before GEMM */
long DBPrec_NNZ_Before(DBPrecMem* mem)
{
  need(mem->L);
  need(mem->E_DB);

  return mem->L + mem->E_DB;
}

/* Sans Gestion Pic : */

/* Memory pic (in nnz) during preconditionner computation WITHOUT pic reduction in GEMM */
/* Pendant GEMM */
long DBPrec_NNZ_In(DBPrecMem* mem)
{
  need(mem->L);
  need(mem->E_DB);
  need(mem->SL);

  return mem->L + mem->E_DB + mem->SL;
}

/* apres GEMM et conversion en phidal si dropping */
long DBPrec_NNZ_After(int Pschur_method, DBPrecMem* mem)
{
  need(mem->L);
  need(mem->SL);
  need(mem->LS);

  if (Pschur_method == 1)
    return mem->L + mem->SL   +    mem->LS;
  else 
    return mem->L + /* mem->SL+ */ mem->LS;
}

/* Avec Gestion Pic / LL */

/* Memory pic (in nnz) during preconditionner computation when we
   desallocate EU^-1 and allocate S in GEMM */
long DBPrec_NNZ_LL(DBPrecMem* mem)
{

  need(mem->L);
  need(mem->ll);

  return mem->L + mem->ll;
}

/* Avec Gestion Pic / RL */

/* TODO : réduire le nb de fonction */
long DBPrec_NNZ_RL(DBPrecMem* mem)
{

  need(mem->L);
  need(mem->SL);
  need(mem->Emax);

  return mem->L + mem->SL + mem->Emax;
}

long DBPrec_NNZ_RL_DROPE(DBPrecMem* mem)
{

  need(mem->L);
  need(mem->dropE);
  need(mem->Emax);

  return mem->L + mem->dropE + mem->Emax;
}

long DBPrec_NNZ_RL_DROPEAfter(DBPrecMem* mem)
{
  /* TODO : ajouter S */
  need(mem->L);
  need(mem->dropEAfter);

  return mem->L + mem->dropEAfter;
}
