/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBDistrMatrix_solve_sub.c                                  */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>

#include <string.h> /* memcpy */

#include "block.h"
#include "db_parallel.h"

void TR(flag_t unitdiag, SolverMatrix* solvUs, SymbolMatrix* symbUs, COEF *x, COEF *b, COEF* xtmp);

void TR1(flag_t unitdiag, VSolverMatrix* vsolvmtx, COEF* b, COEF* x, COEF* xtmp) {
  TR(unitdiag, vsolvmtx->solvmtx, &vsolvmtx->symbmtx, b, x, xtmp);
}

void TR(flag_t unitdiag, SolverMatrix* solvUs, SymbolMatrix* symbUs, COEF *x, COEF *b, COEF* xtmp) {
  dim_t k, p;
  int jxtmp;

  dim_t tli = symbUs->tli; /* cf LSolv */

  COEF* xptr;
  int height, width, stride;

  int size;

  COEF minusone= -1.0, one=1.0;
  char *uploL = "L";
  char *transT = "T";
  char* diag = unitdiag?"U":"N";
  int UN = 1;
  assert((unitdiag == 0) || (unitdiag == 1));
  
  for(k=symbUs->cblknbr-1; k>=0; k--) {
    width  = symbUs->ccblktab[k].lcolnum - symbUs->ccblktab[k].fcolnum +1;
    xptr = x + symbUs->ccblktab[k].fcolnum - tli;
      	
    /* xtmp := x (in) */
     
    jxtmp=0;
    
    p = symbUs->bcblktab[k].fbloknum+1;
    if (p> symbUs->bcblktab[k].lbloknum) goto next;
	
    for(; p<=symbUs->bcblktab[k].lbloknum; p++) {
      size = symbUs->bloktab[p].lrownum - symbUs->bloktab[p].frownum +1;

      BLAS_COPY(size, x + symbUs->bloktab[p].frownum - tli, UN, xtmp + jxtmp, UN);
      jxtmp += size;
    }

   
    p = symbUs->bcblktab[k].fbloknum+1;
    height = symbUs->hdim[k] - width; /** Hauteur de la surface compactÃ©e des blocs extra-diagonaux **/    
    stride = symbUs->stride[k];
    if (height == 0) goto next;   /* inutile ? */
        	
    /* calcul des contributions (aggrégation des colonnes) */
    BLAS_GEMV(transT, height, width, minusone /*coef alpha*/,
	      solvUs->coeftab + solvUs->bloktab[p].coefind, stride,
	      xtmp, UN,       one /*coef beta*/,
	      xptr, UN);
	
  next:
    stride = symbUs->stride[k];
    p = symbUs->bcblktab[k].fbloknum; /* triangular block */
    BLAS_TRSV(uploL, transT, diag, width /*=h*/,
	      solvUs->coeftab + solvUs->bloktab[p].coefind, stride, xptr, UN);
  }
}

/* void TR(flag_t unitdiag, SolverMatrix* solvUs, SymbolMatrix* symbUs, COEF *x, COEF *b, COEF* xtmp) { */
/*   dim_t k, p; */
/*   int jxtmp; */

/*   dim_t tli; */
/*   printfv(5, "TODO : tli !\n"); */
/* /\*   if (U->alloc != BLK) *\/ */
/* /\*     tli = U->a[0].symbmtx.tli; *\/ */
/* /\*   else /\\* (U->alloc == BLK) *\\/ *\/ */
/* /\*     tli = U->ca[0]->symbmtx.tli; *\/ */

/*   COEF* xptr; */
/*   int height, width, stride; */

/*   int size; */

/*   COEF minusone= -1.0, one=1.0; */
/*   char *uploL = "L"; */
/*   char *transT = "T"; */
/*   char* diag = unitdiag?"U":"N"; */
/*   int UN = 1; */
/*   assert((unitdiag == 0) || (unitdiag == 1)); */
 	
/*   for(k=symbUs->cblknbr-1; k>=0; k--) { */
/*     width  = symbUs->ccblktab[k].lcolnum - symbUs->ccblktab[k].fcolnum +1; */
/*     xptr = x + symbUs->ccblktab[k].fcolnum - tli; */
      	
/*     /\* xtmp := x (in) *\/ */
     
/*     jxtmp=0; */
    
/*     p = symbUs->bcblktab[k].fbloknum+1; */
/*     if (p> symbUs->bcblktab[k].lbloknum) goto next; */
	
/*     for(; p<=symbUs->bcblktab[k].lbloknum; p++) { */
/*       size = symbUs->bloktab[p].lrownum - symbUs->bloktab[p].frownum +1; */
/*       BLAS_COPY(size, x + symbUs->bloktab[p].frownum - tli, UN, xtmp + jxtmp, UN); */
/*       jxtmp += size; */
/*     } */

   
/*     p = symbUs->bcblktab[k].fbloknum+1; */
/*     height = symbUs->hdim[k] - width; /\** Hauteur de la surface compactÃ©e des blocs extra-diagonaux **\/     */
/*     stride = symbUs->stride[k]; */
/*     if (height == 0) goto next;   /\* inutile ? *\/ */
        	
/*     /\* calcul des contributions (aggrégation des colonnes) *\/ */
/*     BLAS_GEMV(transT, height, width, minusone /\*coef alpha*\/, */
/* 	      solvUs->coeftab + solvUs->bloktab[p].coefind, stride, */
/* 	      xtmp, UN,       one /\*coef beta*\/, */
/* 	      xptr, UN); */
	
/*   next: */
/*     stride = symbUs->stride[k]; */
/*     p = symbUs->bcblktab[k].fbloknum; /\* triangular block *\/ */
/*     BLAS_TRSV(uploL, transT, diag, width /\*=h*\/, */
/* 	      solvUs->coeftab + solvUs->bloktab[p].coefind, stride, xptr, UN); */
/*   } */
/* } */

void DBMATVEC2(SolverMatrix* solvUs, SymbolMatrix* symbUs, COEF *x, COEF *b, COEF* xtmp);

void DBMATVEC21(VSolverMatrix* vsolvmtx, COEF* b, COEF* x, COEF* xtmp) {
  DBMATVEC2(vsolvmtx->solvmtx, &vsolvmtx->symbmtx, b, x, xtmp);
}



void DBMATVEC2(SolverMatrix* solvUs, SymbolMatrix* symbUs, COEF *x, COEF *b, COEF* xtmp) {
  dim_t k, p;
  int jxtmp;
  
  dim_t tli = symbUs->tli; /* cf LSolv */

  COEF* xptr;
  int height, width, stride;

  int size;

  COEF minusone= -1.0, one=1.0;
  char *transT = "T";
  int UN = 1;

  for(k=symbUs->cblknbr-1; k>=0; k--) {
    width  = symbUs->ccblktab[k].lcolnum - symbUs->ccblktab[k].fcolnum +1;
    xptr = x + symbUs->ccblktab[k].fcolnum - tli;
      	
    /* xtmp := x (in) */
     
    jxtmp=0;
    
    p = symbUs->bcblktab[k].fbloknum;
    if (p> symbUs->bcblktab[k].lbloknum) continue;
	
    for(; p<=symbUs->bcblktab[k].lbloknum; p++) {
      size = symbUs->bloktab[p].lrownum - symbUs->bloktab[p].frownum +1;

      BLAS_COPY(size, x + symbUs->bloktab[p].frownum - tli, UN, xtmp + jxtmp, UN);
      jxtmp += size;
    }

   
    p = symbUs->bcblktab[k].fbloknum;
    height = symbUs->hdim[k]; /** Hauteur de la surface compactÃ©e des bloc extra-diagonaux **/
    stride = symbUs->stride[k];
    if (height == 0) continue;
   	
    /* calcul des contributions (aggrégation des colonnes) */
    BLAS_GEMV(transT, height, width, minusone /*coef alpha*/,
	      solvUs->coeftab + solvUs->bloktab[p].coefind, stride,
	      xtmp, UN,       one /*coef beta*/,
	      xptr, UN);
  }
}

