/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBDistrPrec_SchurProd.c                                    */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#define PARALLEL

#include "block.h"
#include "db_parallel.h"
#include "phidal_parallel.h"

/* todo : uniquement sÃ©quentiel, juste pour le prototype */

void DBDistrPrec_SchurProd(DBDistrPrec *P, PhidalDistrHID *DBL, COEF *x, COEF *y)
{
  /**************************************************************/
  /* This function compute the implicit Schur Product           */
  /* y = S.x    where S is not stored                           */
  /* eg y = B.x - E.U^-1.L^1.F.x (for unsymmetric)              */
  /**************************************************************/

  /*** @@OIMBE PROBLEM SI PERMUTATION ????? B n'EST PAS PERMUTE ****/

  PhidalHID* BL = &DBL->LHID;

  COEF *t;
  
  DBMatrix *L, *U;
  PhidalDistrMatrix *E, *F;
  PhidalDistrMatrix *B;

#ifdef DEBUG_M
  assert(P->schur_method == 2);
#endif

  E = PREC_E(P);
  F = PREC_F(P);
  L = S(PREC_L_BLOCK(P));
  U = S(PREC_U_BLOCK(P));
  B = PREC_B(P);

  t = (COEF *)malloc(sizeof(COEF)*L->dim1);

  PhidalDistrMatrix_MatVec(0, B, DBL, x, y);
  PhidalDistrMatrix_MatVec(0, F, DBL, x, t);

  DBMatrix_Lsolv(1, L, t, t, BL);
  if(P->symmetric == 1)
    DBMatrix_Dsolv(L, t);
  DBMatrix_Usolv(P->symmetric, U, t, t, BL);

  PhidalDistrMatrix_MatVecSub(0, E, DBL, t, y);

  free(t);
}
