/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : DBDistrPrec_func.c                                         */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#define PARALLEL

#include "block.h"
#include "db_parallel.h"
#include "phidal_parallel.h"

void DBDistrPrec_Unscale(REAL *scaletab, REAL *iscaletab, DBDistrPrec *P, PhidalDistrHID *BL)
{

  /*****************************************************/
  /*  A = Dr^-1.A'.Dr^-1                               */
  /*  A = Dr^-1.L'.D'.L't.Dr^-1                        */
  /*  A = Dr^-1.L'.Dr . Dr^-1.D'.Dr^-1  . Dr.L't.Dr^-1 */
  /*  A =     L       .      D          . Lt           */
  /*****************************************************/

#ifdef DEBUG_M
  assert(P->symmetric == 1);
#endif

  DBDistrMatrix_RowMult(iscaletab, PREC_L_BLOCK(P), BL);  
  DBDistrMatrix_ColMult(scaletab, PREC_L_BLOCK(P), BL);
  
  DBDistrMatrix_DiagMult(iscaletab, PREC_L_BLOCK(P), BL); /* iscaletab for DB, scaletab for Phidal */
  
  if(P->forwardlev > 0)
    {
      
      if (P->nextprec != NULL) {
        DBDistrPrec_Unscale(scaletab + PREC_L_BLOCK(P)->M.dim1, iscaletab + PREC_L_BLOCK(P)->M.dim1, P->nextprec, BL);
      } else {
        PhidalDistrPrec_SymmetricUnscale(scaletab + PREC_L_BLOCK(P)->M.dim1, iscaletab + PREC_L_BLOCK(P)->M.dim1, P->phidalPrec, BL);
      }

      /* Si plus de deux niveaux*/
      /*       if(P->E->virtual == 0) */
      /*        { */
      /*          PhidalDistrMatrix_RowMult(iscaletab + PREC_L_BLOCK(P)->dim1, P->E, BL); */
      /*          PhidalDistrMatrix_ColMult(iscaletab, P->E, BL); */
      /*        } */
 
      if(P->schur_method == 1)
        {
          if (PREC_SL_BLOCK(P) != NULL) { /* equivalent (option->droptol1 == 0) (& -1 hack) */
            DBDistrMatrix_RowMult(iscaletab + PREC_L_BLOCK(P)->M.dim1, PREC_SL_BLOCK(P), BL);  
            DBDistrMatrix_ColMult(iscaletab + PREC_L_BLOCK(P)->M.dim1, PREC_SL_BLOCK(P), BL);
          } else {
            PhidalDistrMatrix_RowMult2(iscaletab + PREC_L_BLOCK(P)->M.dim1, P->phidalS, BL);
            PhidalDistrMatrix_ColMult2(iscaletab + PREC_L_BLOCK(P)->M.dim1, P->phidalS, BL);
          }


	}

      
    }
 
}

void DBDistrPrec_Unscale_Unsym(REAL *scalerowtab, REAL *iscalerowtab, REAL *scalecoltab, REAL *iscalecoltab, DBDistrPrec *P, PhidalDistrHID *BL)
{

  COEF* D;

  DBDistrMatrix_RowMult(iscalerowtab, PREC_L_BLOCK(P), BL);
  DBDistrMatrix_ColMult(scalerowtab,  PREC_L_BLOCK(P), BL);
  
  DBDistrMatrix_Transpose(PREC_U_BLOCK(P));
  DBDistrMatrix_RowMult(iscalecoltab, PREC_U_BLOCK(P), BL); /* /!\ doit Ãªtre inversÃ© par rapport Ã  Phidal (Transpose) */
  DBDistrMatrix_ColMult(iscalerowtab, PREC_U_BLOCK(P), BL); /* /!\ doit Ãªtre inversÃ© par rapport Ã  Phidal (Transpose) */
  DBDistrMatrix_Transpose(PREC_U_BLOCK(P));

  /* ? */
  D = (COEF *)malloc(sizeof(COEF)*PREC_L_BLOCK(P)->M.dim1);
  DBDistrMatrix_Transpose(PREC_U_BLOCK(P));
  DBMatrix_GetDiag(&PREC_U_BLOCK(P)->M, D);
  DBDistrMatrix_Transpose(PREC_U_BLOCK(P));
  DBMatrix_SetDiag(&PREC_L_BLOCK(P)->M, D);
  free(D);

  if(P->forwardlev > 0)
    {
      
      if (P->nextprec != NULL) {
        DBDistrPrec_Unscale_Unsym(scalerowtab + PREC_L_BLOCK(P)->M.dim1, iscalerowtab + PREC_L_BLOCK(P)->M.dim1, 
                                  scalecoltab+PREC_U_BLOCK(P)->M.dim1, iscalecoltab+PREC_U_BLOCK(P)->M.dim1, P->nextprec, BL);
      } else {
        PhidalDistrPrec_UnsymmetricUnscale(scalerowtab + PREC_L_BLOCK(P)->M.dim1, iscalerowtab + PREC_L_BLOCK(P)->M.dim1, 
                                           scalecoltab+PREC_U_BLOCK(P)->M.dim1, iscalecoltab+PREC_U_BLOCK(P)->M.dim1, P->phidalPrec, BL);
      }
      
      if(P->schur_method == 1)
        {
          if (PREC_SL_BLOCK(P) != NULL) { /* equivalent (option->droptol1 == 0) (& -1 hack) */
            DBDistrMatrix_RowMult(iscalerowtab + PREC_L_BLOCK(P)->M.dim1, PREC_SL_BLOCK(P), BL);
            DBDistrMatrix_ColMult(iscalecoltab + PREC_L_BLOCK(P)->M.dim1, PREC_SL_BLOCK(P), BL);
            DBDistrMatrix_RowMult(iscalecoltab + PREC_U_BLOCK(P)->M.dim1, PREC_SU_BLOCK(P), BL);
            DBDistrMatrix_ColMult(iscalerowtab + PREC_U_BLOCK(P)->M.dim1, PREC_SU_BLOCK(P), BL);
          } else {
            PhidalDistrMatrix_RowMult2(iscalerowtab + PREC_L_BLOCK(P)->M.dim1, P->phidalS, BL);
            PhidalDistrMatrix_ColMult2(iscalecoltab + PREC_L_BLOCK(P)->M.dim1, P->phidalS, BL);
          }
        }


    }
}
