/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : VS_RowDiv.c                                                */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <math.h>
#include <stdio.h>
#include <stdlib.h>

#include <assert.h>
#include "block.h"

#include "base.h"

void VS_RowDiv(VSolverMatrix* D, VSolverMatrix* M) {
  SolverMatrix* solvD = D->solvmtx;
  SolverMatrix* solvM = M->solvmtx;

  SymbolMatrix* symbD = &D->symbmtx;
  SymbolMatrix* symbM = &M->symbmtx;

  dim_t k, m;
  COEF alphadiag;
  blas_t strideD, strideM;
  int cdim, hdim;
  COEF *dc; 
  COEF *mc; 

  int UN = 1;

  /* ./testDBDistrMatrix.ex 7 bcsstk01.rsa */
  if (symbM->bloknbr == 0) return;

  for(k=0;k<symbM->cblknbr;k++)
    {
      cdim    = symbM->ccblktab[k].lcolnum - symbM->ccblktab[k].fcolnum + 1;
      strideD = symbD->stride[k];
      strideM = symbM->stride[k];
      hdim    = symbM->hdim[k];

      dc = solvD->coeftab + solvD->bloktab[ symbD->bcblktab[k].fbloknum ].coefind;
      mc = solvM->coeftab + solvM->bloktab[ symbM->bcblktab[k].fbloknum ].coefind;

      for(m=0;m<cdim;m++) {
#ifndef DEBUG_NOALLOCATION      
	alphadiag = 1.0 / dc[m*strideD + m];
#endif
	BLAS_SCAL(hdim, alphadiag, mc, UN);
	mc += strideM;
      }
      
    }
  
}

