/******************************************************************************/
/* Copyright 2008-2009 CNRS                                                   */
/*                                                                            */
/* File          : SolverMatrix_Div.c                                         */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR                                               */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#ifdef OBSOLETE
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>

#include "phidal_struct.h"

#include "solver.h"

#include "base.h"

void SolverMatrix_Div(SolverMatrix* A, SolverMatrix* D) {
  /************************************************************************************/
  /* This function performs A = A.D^1   (D diag)                                      */
  /************************************************************************************/

  SymbolMatrix *symbA = &(A->symbmtx);
  SymbolMatrix *symbD = &(D->symbmtx);

  dim_t k,m;
  int cdim, rdim;
  blas_t strideA, strideD;

  REAL *ac, *dc;

  REAL alpha;
  int UN = 1;

  /* Diviser A par la diagonale de D */
  for(k=0;k<symbA->cblknbr;k++)
    {
      cdim = symbA->ccblktab[k].lcolnum - symbA->ccblktab[k].fcolnum +1; /** Largeur du bloc colonne **/
      rdim = symbA->hdim[k]; /* the number of rows of the matrix A */

      strideA = symbA->stride[k];
      strideD = symbD->stride[k];

      ac = A->coeftab + A->bloktab[ symbA->bcblktab[k].fbloknum ].coefind;
      dc = D->coeftab + D->bloktab[ symbD->bcblktab[k].fbloknum ].coefind;

      for(m=0;m<cdim;m++) {
	alpha = 1.0/ dc[m*strideD + m]; /** alpha est l'inverse du m-ieme terme diagonal du bloc diagonal de D **/
	DSCAL(rdim, alpha, ac, UN); /** On divise la m-ieme colonne par le terme diagonal **/
	ac += strideA;
      }
    }
}


/*TODO : logique dans += ou calcul a chaque fois*/
/*TODO : changer SOPLAIN*/
#endif
