/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : hips_error.c                                               */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <limits.h>
#include <mpi.h>
#include "math.h"

#include "hips.h"
#include "hips_wrapper.h"
#include "scotch_metis_wrapper.h"

void HIPS_PrintError(INTS ierr)
{
  switch(ierr)
    {
    case HIPS_SUCCESS:
      break;
    case HIPS_ERR_ALLOCATE :
      fprintferr(stderr, "ERROR %ld : failed a memory allocation \n", (long)ierr);
      break;
    case HIPS_ERR_PARAMETER :
      fprintferr(stderr, "ERROR %ld : a parameter is wrong or a function is called at a wrong place \n", (long)ierr);
      break;
    case HIPS_ERR_CALL :
      fprintferr(stderr, "ERROR %ld : a function is called at a wrong place \n", (long)ierr);
      break;
    case HIPS_ERR_IO : 
      fprintferr(stderr, "ERROR %ld : IO problem on a file \n", (long)ierr);
      break;
    case HIPS_ERR_MATASSEMB :
      fprintferr(stderr, "ERROR %ld : problem in when setting the matrix coefficients \n", (long)ierr);
      break;
    case HIPS_ERR_RHSASSEMB :
      fprintferr(stderr, "ERROR %ld : problem in when setting the rhs coefficients \n", (long)ierr);
      break;
    case HIPS_ERR_PARASETUP :
      fprintferr(stderr, "ERROR %ld : problem in the parallel setup phase \n", (long)ierr);
      break;
    case HIPS_ERR_PRECOND :
      fprintferr(stderr, "ERROR %ld : problem in the preconditioner building \n", (long)ierr);
      break;
    case HIPS_ERR_SOLVE :
      fprintferr(stderr, "ERROR %ld : problem in the solve phase \n", (long)ierr);
      break;
    case HIPS_ERR_KRYLOV :
      fprintferr(stderr, "ERROR %ld : the krylov method failed \n", (long)ierr);
      break;
    case HIPS_ERR_CHECK :
      fprintferr(stderr, "ERROR %ld : a check point failed \n", (long)ierr);
      break;
    default:
      fprintferr(stderr, "ERROR in HIPS_PrintError : error %ld is not referenced \n", (long)ierr);
    }

}

#ifdef DEBUG_M
#define MY_EXIT assert(0)
#else
#define MY_EXIT exit(-1)
#endif

void HIPS_ExitOnError(INTS ierr)
{

  HIPS_PrintError(ierr);
  if(ierr != HIPS_SUCCESS)
    MY_EXIT;

}

