/******************************************************************************/
/* Copyright 2008-2009 CNRS, INRIA                                            */
/*                                                                            */
/* File          : read_rhs.c                                                 */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR and P. HENON                                  */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>

#include "phidal_common.h"

#define BUFLEN 200

void VECread(char *filename, INTS n, COEF *rhs)
{
  /***********************************************/
  /* Read a vector from the file filename        */
  /* num=1 Fortran numbering                     */
  /* rhs must have been allocated with at least  */
  /* n element                                   */
  /***********************************************/

  FILE *fp;
  char buf[BUFLEN];
  int i;/* , ind; */
  double tmp1;
#ifdef TYPE_COMPLEX
  double tmp2;
#endif

  if((fp = fopen(filename,"r")) == NULL) 
    {
      fprintfd(stderr, "read_rhs: Cannot open file %s \n", filename);
      exit(1);
    }
  i = 0;
  while(fgets(buf, BUFLEN, fp) != NULL) 
    {
      if(i>= n)
	{
	  fprintf(stderr, "WARNING : file %s contains a vector of dimension > n = %ld \n", filename, (long)n);
	  break;
	}
      /*assert(i<n);*/

      /*sscanf(buf, "%d %lf", &ind, &tmp1); Pour IFP **/
#ifdef TYPE_REAL
      sscanf(buf, "%lf", &tmp1);
      rhs[i] = tmp1;
#else
      sscanf(buf, "%lf %lf", &tmp1, &tmp2);
      rhs[i] = tmp1 + I*tmp2;
#endif
      i++;
    }

  if(i<n)
    {
      fprintf(stderr, "ERROR : file %s contains a vector of dimension %ld < n = %ld \n", filename, (long)i, (long)n);
      assert(n == i);
    }
  fclose(fp);

  /*fprintfv(5, stderr, "nnz = %d \n", i);*/
}
