/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : CSCrowICCprod.c                                            */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <stdio.h>

#include "phidal_common.h"


void CSCrowICCprod(REAL droptol, REAL *droptab, REAL fillrat, dim_t nnb, csptr *X, csptr l, COEF *D, int *wki1, int *wki2, COEF *wkd, cell_int *celltab, cell_int **cellptrtab)
{
  /*************************************************************/
  /* This function performs l = l - X.D.Xt                     */
  /* X is a row of CSC matrices                                */
  /* l is a CSC lower triangular matrix                        */
  /*                                                           */
  /* On return:                                                */
  /* l the result                                              */
  /* X = X.D and X is sorted by row indices                    */
  /*************************************************************/
  csptr x;

  if(nnb == 0)
    return;
  
  /**** We compact X in a single CSC matrix ******/
  x = (csptr)malloc(sizeof(struct SparRow));
  CS_VirtualMerge(nnb, X, x);

  CS_ICCprod(x, l, D, droptol, droptab, fillrat, wki1, wki2, wkd, celltab, cellptrtab); /** UPDATED ROW STRUCT IS HERE **/

  /** Reinit the pointer of x and y to no delete them in cleanCS **/
  /*bzero(x->nnzrow, sizeof(int)*x->n);
    bzero(x->ja, sizeof(int *)*x->n);
    bzero(x->ma, sizeof(COEF *)*x->n);
    cleanCS(x);*/

    /** Destroy x : DO NOT USE cleanCS (it uses a lot of time in this case) **/
  free(x->nnzrow);
  free(x->ja);
  free(x->ma);
  free(x->nzrtab);
  free(x->inzrtab);
  free(x);
}
