/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : iluk.c                                                     */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <assert.h>
#include "phidal_common.h"
#include "queue.h"

int LDLk(csptr L, REAL normA, REAL epsilon)
{
  /******************************************************************************************/
  /* This function computes the LDLt(k) factorization of a sparse matrix                    */
  /* We assume that the sparse pattern of the matrix has been symbolically                  */
  /* factorized (the non zero pattern of the final factor has already been allocated        */
  /* == call to symb_fact)                                                                  */
  /* On entry :                                                                             */
  /*  L : a csc matrix that contains the lower triangular part of the matrix                */
  /* normA : the norm of matrix (used for static numerical pivoting)                        */
  /* epsilon : the tolerance that will be used in the accelerator (also used for pivoting)  */
  /* On return :                                                                            */
  /*  L : the factorized matrix                                                             */
  /*  the diagonal stores the inverse of the factorized matrix diagonal                     */
  /* NOTE:                                                                                  */
  /* This function can use any sparse pattern it is not optimized for a direct factorization*/
  /******************************************************************************************/


  /*** @@@@@@@@@@ OIMBE DOIT POUVOIR MIEUX CODER l'ajout des contribution (voir sparseTRSM)
       puisque on sait que l'on peut toujours ajouter dans la colonne distante EN DIRECT 
       l'algo actuelle peut traiter incomplete et complete factorization sparse pattern ****/

  dim_t j, k;
  int indi, indj;
  dim_t *jaColi, *jaColj;
  COEF *maColi, *maColj;
  int *Ljaptr;
  COEF *Lmaptr;
  dim_t nnzrow;
  int fnnzrow;
  REAL d;
  REAL thresh;

  /** Threshold for pivoting **/
  thresh = sqrt(epsilon)*normA;

  for(k=0;k<L->n;k++)
    {
      Ljaptr = L->ja[k];
      Lmaptr = L->ma[k];
      nnzrow = L->nnzrow[k];

      /** Static pivoting **/
      if(fabs(Lmaptr[0]) < thresh)
	{
	  fprintfd(stderr, "Pivot \n");
	  Lmaptr[0] = thresh;
	}
	
      Lmaptr[0] = 1.0/Lmaptr[0];
      

      /** Update lower triangular part L **/
      for(j=1;j<nnzrow;j++)
	{
	  jaColi = L->ja[ Ljaptr[j] ];
	  maColi = L->ma[ Ljaptr[j] ];
	  jaColj = Ljaptr + j;
	  maColj = Lmaptr + j;

	  fnnzrow = L->nnzrow[ (*jaColj) ];

	  d = -(*maColj)*Lmaptr[0];

	  indj = j;
	  indi = 0;
	  while(indi < fnnzrow && indj < nnzrow)
	    {
	      
	      while(indi < fnnzrow && (*jaColi) < (*jaColj) )
		{
		  jaColi++;
		  maColi++;
		  indi++;
		}

	      if(indi >= fnnzrow)
		break;

	      while(indj < nnzrow && (*jaColi) > (*jaColj) )
		{
		  jaColj++;
		  maColj++;
		  indj++;
		}
	      
	      if(indj >= nnzrow)
		break;
	      
	      if( (*jaColi) !=  (*jaColj))
		continue;
	      
#ifdef DEBUG_M
	      assert((*jaColi) ==  (*jaColj));
#endif


#ifndef WORSE
	      /** Much Better solution **/
	      while(indi < fnnzrow && indj < nnzrow && (*jaColi) ==  (*jaColj))
		{
		  (*maColi) += d* (*maColj);
		  jaColj++;
		  maColj++;
		  jaColi++;
		  maColi++;
		  indi++;
		  indj++;
		}
#else /** Worse **/
	      cur = 0;
	      while( (indi+cur < fnnzrow && indj+cur < nnzrow) && (jaColi[cur] ==  jaColj[cur]) )
		cur++;

	      DAXPY(cur, d, maColj, incx, maColi, incx);
	      
	      if(indi+cur < fnnzrow && indj+cur < nnzrow)
		{
		  indi+=cur;
		  indj+=cur;
		  jaColj += cur;
		  maColj += cur;
		  jaColi += cur;
		  maColi += cur;
		}
	      else
		break;
#endif
	    }

	  Lmaptr[j] = -d;
	}
      
    }

  CS_SetNonZeroRow(L);
  return 1;
}
