/******************************************************************************/
/* Copyright 2008-2009 CNRS, INRIA                                            */
/*                                                                            */
/* File          : Improve_Partition.c                                        */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR and P. HENON                                  */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include "phidal_ordering.h"

void Improve_Partition(flag_t verbose, dim_t tagnbr, INTL n, INTL *ig, dim_t *jg, dim_t **mapp, dim_t **mapptr, dim_t *ndom, PhidalHID *BL, dim_t *perm, dim_t *iperm)
{
  /**********************************************************************/
  /* This function recompute a new HID from the previous one            */
  /* Principle is to take a point in each connector of the last level : */
  /* these points are the cross point of a new domain partition         */
  /**********************************************************************/
  dim_t *nodelist;
  dim_t i, nodenbr, level;
  int fillopt[10];

#ifdef DEBUG_M
  assert(tagnbr > 0 && tagnbr <= n);
#endif



  fprintfd(stderr, "IMPROVE START \n");
  /*HID_Info(stdout, BL);*/

  level = BL->nlevel-1;
  for(;level >= 0;level--)
    {
      fprintfd(stderr, "level = %ld points nbr = %ld \n", (long)level, (long)(BL->nblock - BL->block_levelindex[level]));
      if(BL->block_levelindex[BL->nlevel] - BL->block_levelindex[level] >= (*ndom)/tagnbr)
	break;
      if(level > 0 && (BL->block_levelindex[BL->nlevel] - BL->block_levelindex[level-1] >= 2*(*ndom)/tagnbr) )
	break;
      
    }
  
  fprintfd(stderr, "level = %ld in Improve_Partition : we use %ld points \n", (long)level, (long)(BL->block_levelindex[BL->nlevel] - BL->block_levelindex[level]));
  
  nodelist = (dim_t *)malloc(sizeof(dim_t)*(BL->nblock-BL->block_levelindex[level]));
  nodenbr = 0;
  for(i=BL->block_levelindex[level];i<BL->nblock;i++)
    nodelist[nodenbr++] = iperm[BL->block_index[i]];
  


  /** APPEL FONCTION CONSTRUCTION DOMAIN X ***/


  free(nodelist);

  /*** Recompute the HID ***/
  PHIDAL_HierarchDecomp(verbose, 0, n, ig, jg, *mapp, *mapptr, *ndom, BL, perm, iperm);
  

  /*** Reorder the interior domain ****/
  CSR_Perm(n, ig, jg, NULL, perm);
  fillopt[0] = 0; /** 0 ND, 1 MD, 2 MF **/

  PHIDAL_MinimizeFill(fillopt, n, ig, jg, BL, iperm);
  for(i=0;i<n;i++)
    perm[iperm[i]] = i;
}
