/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PHIDAL_Interior2OverlappedPartitions.c                     */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "phidal_ordering.h"
#include "scotch_metis_wrapper.h"

void PHIDAL_Interior2OverlappedPartition(int numflag, int ndom, int n, INTL *ia, dim_t *ja, int *node2dom, int **mapp, int **mapptr)
{
  /*------------------------------------------------------------------------------------/
  /   This function tranforms an interior domain partition in an overlapped partition  /
  
  /   On entry:                                                                         /
  /   numflag    : =0 the matrix is given in C numbering                               /
  /                 =1 the matrix is given in Fortran numbering                         /
  /   n, ia, ja   : matrix pattern                                                      /
  /   node2dom    : node2dom[i] is the subdomain number that contains node i            /
  /                                                                                     /
  /   On return:                                                                        /
  /   mapp, mapptr: the edge based partition (overlapped partition)                     /
  /                                                                                     /
  /  NOTE: input matrix must be in C numbering (indices start from 0)                   /
  /------------------------------------------------------------------------------------*/
  dim_t i, j, k;
  int ind;
  int *mapptr1;
  int *mapp1; 
  

   if(numflag == 1)
     CSR_Fnum2Cnum(ja, ia, n);


  mapp1 = (int *)malloc(sizeof(int)*n);
  mapptr1 = (int*)malloc(sizeof(int)*(ndom+1));
  bzero(mapptr1, sizeof(int)*(ndom+1));
  
  ind = 0;
  for(k=0;k<ndom;k++)
    for(i=0;i<n;i++)
      if(node2dom[i] == k)
	{
	  mapp1[ind++] = i;
	  mapptr1[k+1]++;
	}

#ifdef DEBUG_M
  assert(ind <= n);
#endif

  for(k=0;k<ndom;k++)
    mapptr1[k+1] += mapptr1[k];
  
#ifdef DEBUG_M  
  assert(mapptr1[ndom] <= n);
#endif
  *mapptr = mapptr1;
  *mapp = mapp1;

#ifdef DEBUG_M
  /*** Check if any interior node of a subdomain is related to another subdomain interior node ***/
  fprintfd(stderr, "Check interior domains \n");
  check_interior_node(ndom, (*mapptr), (*mapp), ia, ja, n);
  fprintfd(stderr, "Check interior domains DONE \n");
#endif

  
  /**** Expand domain to get overlap ****/
 
  get_overlap(ndom, n, mapptr, mapp, ia, ja);
  
   if(numflag == 1)
     CSR_Cnum2Fnum(ja, ia, n);
}
