/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : vec_balanced_partition.c                                   */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <string.h>
#include "queue.h"


int vec_balanced_partition(int ndom, int n, REAL *W, int *elt2dom)
{
  /******************************************************************/
  /* This function partition contiguously a 1 dimensionnal domain   */
  /* composed of n insecable component. The weighted of each        */
  /* element is W[i]. In return elt2dom[i] is the domain of the     */
  /* component i                                                    */
  /******************************************************************/
  int i, dnum;
  REAL Wmean, Wloc, Wideal;

  if(ndom>n)
    return -1; /** Not able to partition **/
#ifdef DEBUG_M
  assert(ndom > 0);
  for(i=0;i<n;i++)
    assert(W[i]>0);
#endif

  Wmean = 0.0;
  for(i=0;i<n;i++)
    Wmean += W[i];
  Wmean /= ndom;

  Wloc = 0.0;
  i = 0;
  for(dnum=0;dnum<ndom;dnum++)
    {
      /** We force at least one element in the domain i **/
      Wloc += W[i];
      elt2dom[i] = dnum;
      i++;

      Wideal = Wmean*(dnum+1);
      while(i < n-ndom+dnum+1 && Wideal > Wloc+W[i]/2.0 )
	{
	  elt2dom[i] = dnum;
	  Wloc += W[i]; 
	  i++;
	}
    }
#ifdef DEBUG_M
  assert(i == n);
#endif
    
  return EXIT_SUCCESS;
}
