/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalDistrMatrix_Build.c                                  */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "phidal_parallel.h"


void PhidalDistrMatrix_Build(flag_t symmetric, char *UPLO, csptr A, PhidalDistrMatrix *DM, PhidalDistrHID *DBL)
{
  /****************************************************************************/
  /* This function build a phidal matrix from a csr matrix                    */
  /* On entry:                                                                */
  /*   symmetric : if =1 the matrix M will be a CSC lower triangular          */
  /*                Phidal Matrix                                             */
  /*   UPLO: "L": build the lower block triangular matrix  from a             */
  /*         "U": build the upper block triangular matrix from a              */
  /*         "N": build the whole matrix                                      */
  /*   This parameter is usefull for symmetric matrix you need only one half  */
  /*   a   : the matrix you want to transform into PHIDAL matrix              */
  /*   BL  : the phidal hierarchical decomposition structure                  */
  /* On return:                                                               */
  /*   m   : the matrix a in PHIDAL form                                      */
  /****************************************************************************/
  PhidalHID *BL;
  PhidalMatrix *M;
  BL = &DBL->LHID;
  M = &DM->M;


  if(symmetric == 1)
    {
      if(strcmp(UPLO, "U") == 0)
	UPLO = "L";
      else
	{
	  if(strcmp(UPLO, "L") == 0)
	    UPLO = "U";
	  else
	    {
	      fprintfd(stderr, "PhidalMatrix_Build Error : if symmetric == 1 then UPLO must be L or U \n");
	      exit(-1);
	    }
	}
      PhidalDistrMatrix_Setup(0, 0, BL->nblock-1, BL->nblock-1, UPLO, "N", 0, DM, DBL);
      PhidalMatrix_CsrCopy(0, A, "U", M, BL);
#ifdef DEBUG_M
      /*fprintfd(stderr, "Check Before Transpose \n");*/
      PhidalDistrMatrix_Check(DM, DBL);
      /*fprintfd(stderr, "Check Before Transpose DONE\n");*/
#endif
      PhidalDistrMatrix_Transpose(DM);
      M->symmetric = 1;
      M->csc = 1;
    }
  else
    {
      PhidalDistrMatrix_Setup(0, 0, BL->nblock-1, BL->nblock-1, UPLO, "N", 0, DM, DBL);
      M->symmetric = 0;
      M->csc = 0;
      PhidalMatrix_CsrCopy(0, A, "N", M, BL);
    }
  

#ifdef DEBUG_M
  PhidalDistrMatrix_Check(DM, DBL);
#endif

}
