/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : check_parallel.c                                           */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_parallel.h"


int PhidalDistrMatrix_Check(PhidalDistrMatrix *DM, PhidalDistrHID *DBL)
{
  int i, j, k, ind;
  PhidalHID *BL;
  PhidalMatrix *M;

  BL = &DBL->LHID;
  M = &DM->M;
  PhidalMatrix_Check(M, BL);
  
  for(i=M->tli; i<=M->bri;i++)
    {
      for(k=M->ria[i];k<M->ria[i+1];k++)
	{
	  ind = DM->rind[k];
	  if(DM->rlead[k] != DM->proc_id)
	    assert(ind >= 0);

	  if(ind >= 0 && (DM->pset_index[ind+1] - DM->pset_index[ind] <= 1))
	    {
	      fprintfd(stderr, "Error: block (%ld %ld) rind = %ld  \n", (long)i, (long)M->rja[k], (long)ind);
	      assert(DM->pset_index[DM->rind[k]+1] - DM->pset_index[DM->rind[k]] > 1);
	    }
	}
    }

  for(j=M->tlj; j<=M->brj;j++)
    {
      for(k=M->cia[j];k<M->cia[j+1];k++)
	{
	  ind = DM->cind[k];
	  if(DM->clead[k] != DM->proc_id)
	    assert(ind >= 0);
	  if(ind >= 0 && (DM->pset_index[ind+1] - DM->pset_index[ind] <= 1))
	    {
	      fprintfd(stderr, "Error: block (%ld %ld) cind = %ld \n", (long)i, (long)M->cja[k], (long)ind);
	      assert(DM->pset_index[ind+1] - DM->pset_index[ind] > 1);
	    }
	}
    }

  


  return 0;
}
