/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PHIDAL_MatVec.c                                            */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_sequential.h"



void PHIDAL_MatVec(PhidalMatrix *M, PhidalHID *BL, COEF *x, COEF *y)
{
  /*****************************************************************/
  /* This function does y = M.x                                    */
  /*****************************************************************/
  dim_t i,j,k;
  int offseti, offsetj;
  COEF *xptr, *yptr;
  int *bind;


  bind = BL->block_index;

  bzero(y,sizeof(COEF)*M->dim1);
  offseti = bind[M->tli];
  offsetj = bind[M->tlj];
  
  if(M->symmetric == 0)
    {
      if(M->csc == 0)
	for(i=M->tli;i<=M->bri;i++)
	  {
	    yptr = y+(BL->block_index[i]-offseti);
	    for(k=M->ria[i];k<M->ria[i+1];k++)
	      {
		j = M->rja[k];
		matvec_add(M->ra[k], x + (bind[j]-offsetj), yptr);
	      }
	  }
      else
	for(i=M->tli;i<=M->bri;i++)
	  {
	    yptr = y+(BL->block_index[i]-offseti);
	    for(k=M->ria[i];k<M->ria[i+1];k++)
	      {
		j = M->rja[k];
		CSC_matvec_add(M->ra[k], x + (bind[j]-offsetj), yptr);
	      }
	  }
    }
  else
    {
      if(M->csc == 0)
	for(i=M->tli;i<=M->bri;i++)
	  {
	    yptr = y+(BL->block_index[i]-offseti);
	    xptr = x+(BL->block_index[i]-offseti);
	    for(k=M->ria[i];k<M->ria[i+1];k++)
	      {
		j = M->rja[k];
		if(j!=i)
		  matvec_add(M->ra[k], x + (bind[j]-offsetj), yptr);
		else
		  NoDiag_matvec_add(M->ra[k], x + (bind[j]-offsetj), yptr);
		
		matvec_add(M->ra[k], xptr, y + (bind[j]-offsetj));
	      }
	  }
      else
	for(i=M->tli;i<=M->bri;i++)
	  {
	    yptr = y+(BL->block_index[i]-offseti);
	    xptr = x+(BL->block_index[i]-offseti);
	    for(k=M->ria[i];k<M->ria[i+1];k++)
	      {
		j = M->rja[k];
		if(j!=i)
		  CSC_matvec_add(M->ra[k], x + (bind[j]-offsetj), yptr);
		else
		  CSC_NoDiag_matvec_add(M->ra[k], x + (bind[j]-offsetj), yptr);
		
		matvec_add(M->ra[k], xptr, y + (bind[j]-offsetj));
	      }
	  }
    }

}

void PHIDAL_MatVecAdd(PhidalMatrix *M, PhidalHID *BL, COEF *x, COEF *y)
{
  dim_t i;
  COEF *t;
  t = (COEF *)malloc(sizeof(COEF)*M->dim1);
  PHIDAL_MatVec(M, BL, x, t);

  for(i=0;i<M->dim1;i++)
    y[i] += t[i];
  free(t);
}


void PHIDAL_MatVecSub(PhidalMatrix *M, PhidalHID *BL, COEF *x, COEF *y)
{
  dim_t i;
  COEF *t;
  t = (COEF *)malloc(sizeof(COEF)*M->dim1);
  PHIDAL_MatVec(M, BL, x, t);

  for(i=0;i<M->dim1;i++)
    y[i] -= t[i];
  free(t);
}




