/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PHIDAL_SEQ.c                                               */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_sequential.h"



void PHIDAL_MatrixBuild(csptr A, PhidalMatrix *M, PhidalHID *BL, PhidalOptions *option)
{
  if(option->symmetric == 0) {
    PhidalMatrix_Build(0, "N", A, M, BL);
  }
  else {
    PhidalMatrix_Build(1, "L", A, M, BL);
  }
}

void PHIDAL_Precond(PhidalMatrix *A, PhidalPrec *P, PhidalHID *BL, PhidalOptions *option)
{
#ifdef DEBUG_M
  assert(A->symmetric == option->symmetric);
#endif
  
  /** Correct the options **/
  PhidalOptions_Fix(option, BL);

  /** Print the options **/
  if(option->verbose > 0)
    PhidalOptions_Print(stdout, option);

  if(A->symmetric == 0)
    PHIDAL_MLILUTPrec(A, P, BL, option);
  else
    PHIDAL_MLICCTPrec(A, P, BL, option);

}


INTS PHIDAL_Solve( PhidalMatrix *A, PhidalPrec *P, PhidalHID *BL, PhidalOptions *option, COEF *rhs, COEF * x, dim_t *itertab, REAL *resnormtab)
{

  /** There is no flexible PCG so the first level is always
      fgmres **/

  if(option->krylov_method != 1 || (option->forwardlev > 0 && option->schur_method != 0))
    {
      CHECK_RETURN(HIPS_Fgmresd_PH_PH(option->verbose, option->tol, option->itmax, A, P, BL, option, rhs, x, option->fd, itertab, resnormtab));
    }
  else
    {
      CHECK_RETURN(HIPS_PCG_PH_PH(option->verbose, option->tol, option->itmax, A, P, BL, option, rhs, x, option->fd, itertab, resnormtab));
    }
  
  /* if(option->verbose > 0) */
  if(itertab != NULL)
    fprintfv(1, stdout, "Number of outer iterations %d \n", itertab[0]);

  return HIPS_SUCCESS;
}

