/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalMatrix_Drop.c                                        */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "phidal_sequential.h"

#ifdef OLD
/* Just drop the small entry in a "phidal col" of a symmetric PhidalMatrix */
void PhidalMatrix_DropTCol_Sym(int j, PhidalMatrix *A, REAL droptol, REAL *dropptr)
{  
  dim_t k;

  for(k=A->cia[j];k<A->cia[j+1];k++)
    if(A->cja[k] != j) /** Not a diagonal matrix **/ 	 
      CS_DropT(A->ca[k], droptol, dropptr); 	 
    else 	 
      CS_DropT_SaveDiag(A->ca[k], droptol, dropptr);
}

/* Just drop the small entry in a "phidal col" of a unsymmetric PhidalMatrix */
/* TODO : a factoriser en passant en paramètre ria ou cia ... */
void PhidalMatrix_DropTCol_Unsym(int i, PhidalMatrix *A, REAL droptol, REAL *dropptr)
{  
  dim_t k;

  for(k=A->ria[i];k<A->ria[i+1];k++)
    if(A->rja[k] != i) /** Not a diagonal matrix **/ 	 
      CS_DropT(A->ra[k], droptol, dropptr); 	 
    else 	 
      CS_DropT_SaveDiag(A->ra[k], droptol, dropptr);
 
}
#endif

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <assert.h>
#include <math.h>

#include "phidal_sequential.h"



void PhidalMatrix_DropT(PhidalMatrix *A, REAL droptol, REAL *droptab, PhidalHID *BL)
{
  /*****************************************************************/
  /* This function drop the small entry in a phidal matrix         */
  /* The function dropps by row when he matrix is not symmetric    */
  /* and by column when it is symmetric (only lower triangular part*/
  /* is stored                                                     */
  /* all entry in row i <= droptol*droptab[i] are dropped          */
  /* NOTE: the diagonal terms are never dropped                    */
  /*****************************************************************/
  dim_t i, j, k;
  REAL *dropptr;

  if(A->symmetric == 0)
    {
      assert(A->csc == 0);
      for(i=A->tli;i<=A->bri;i++)
	{
	  if(droptab != NULL)
	    dropptr = droptab + (BL->block_index[i]- BL->block_index[A->tli]);
	  else
	    dropptr = NULL;

	  for(k=A->ria[i];k<A->ria[i+1];k++)
	    if(A->rja[k] != i) /** Not a diagonal matrix **/
	      CS_DropT(A->ra[k], droptol, dropptr);
	    else
	      CS_DropT_SaveDiag(A->ra[k], droptol, dropptr);
	}
    }
  else
    {
      assert(A->csc == 1);
      for(j=A->tlj;j<=A->brj;j++)
	{
	  if(droptab != NULL)
	    dropptr = droptab + (BL->block_index[j]- BL->block_index[A->tlj]);
	  else
	    dropptr = NULL;
	  
	  for(k=A->cia[j];k<A->cia[j+1];k++)
	    if(A->cja[k] != j) /** Not a diagonal matrix **/
	      CS_DropT(A->ca[k], droptol, dropptr);
	    else
	      CS_DropT_SaveDiag(A->ca[k], droptol, dropptr);
	}
    }

  
}
