/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalMatrix_Norm.c                                        */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_sequential.h"

/** Local functions **/


void PhidalMatrix_RowNormINF(PhidalMatrix *A, PhidalHID *BL, REAL *normtab)
{
  dim_t i, j, k;
  REAL *normptr;

  bzero(normtab, sizeof(REAL)*A->dim1);

  if(A->csc == 0)
    for(i=A->tli;i<=A->bri;i++)
      {
	normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	for(k=A->ria[i];k<A->ria[i+1];k++)
	  CSRrownormINF(A->ra[k], normptr);
      }
  else
    for(i=A->tli;i<=A->bri;i++)
      {
	normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	for(k=A->ria[i];k<A->ria[i+1];k++)
	  CSRcolnormINF(A->ra[k], normptr);
      }
  
  if(A->symmetric == 1)
    {
      if(A->csc == 0)
	for(j=A->tlj;j<=A->brj;j++)
	  {
	    normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	    for(k=A->cia[j];k<A->cia[j+1];k++)
	      CSRcolnormINF(A->ca[k], normptr);
	  }
      else
	for(j=A->tlj;j<=A->brj;j++)
	  {
	    normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	    for(k=A->cia[j];k<A->cia[j+1];k++)
	      CSRrownormINF(A->ca[k], normptr);
	  }
    }

}

void PhidalMatrix_RowNorm1(PhidalMatrix *A, PhidalHID *BL, REAL *normtab)
{
  dim_t i, j, k;
  REAL *normptr;

  bzero(normtab, sizeof(REAL)*A->dim1);
  if(A->csc == 0)
    for(i=A->tli;i<=A->bri;i++)
      {
	normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	for(k=A->ria[i];k<A->ria[i+1];k++)
	  CSRrownorm1(A->ra[k], normptr);
      }
  else
    for(i=A->tli;i<=A->bri;i++)
      {
	normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	for(k=A->ria[i];k<A->ria[i+1];k++)
	  CSRcolnorm1(A->ra[k], normptr);
      }

  if(A->symmetric == 1)
    {
      if(A->csc == 0)
	for(j=A->tlj;j<=A->brj;j++)
	  {
	    normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
#ifdef DEBUG_M
	    assert(A->cja[A->cia[j]] == j);
#endif
	    CSR_NoDiag_colnorm1(A->ca[A->cia[j]], normptr);
	    for(k=A->cia[j]+1;k<A->cia[j+1];k++)
	      CSRcolnorm1(A->ca[k], normptr);
	  }
      else
	for(j=A->tlj;j<=A->brj;j++)
	  {
	    normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
#ifdef DEBUG_M
	    assert(A->cja[A->cia[j]] == j);
#endif
	    CSR_NoDiag_rownorm1(A->ca[A->cia[j]], normptr);
	    for(k=A->cia[j]+1;k<A->cia[j+1];k++)
	      CSRrownorm1(A->ca[k], normptr);
	  }
    }
}


void PhidalMatrix_RowNorm2(PhidalMatrix *A, PhidalHID *BL, REAL *normtab)
{
  dim_t i, j, k;
  REAL *normptr;

  bzero(normtab, sizeof(REAL)*A->dim1);
  if(A->csc == 0)
    for(i=A->tli;i<=A->bri;i++)
      {
	normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	for(k=A->ria[i];k<A->ria[i+1];k++)
	  CSRrownorm2(A->ra[k], normptr);
      }
  else
    for(i=A->tli;i<=A->bri;i++)
      {
	normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	for(k=A->ria[i];k<A->ria[i+1];k++)
	  CSRcolnorm2(A->ra[k], normptr);
      }

  if(A->symmetric == 1)
    {
      assert(A->csc == 1);
      if(A->csc == 0)
	for(j=A->tlj;j<=A->brj;j++)
	  {
	    normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
#ifdef DEBUG_M
	    assert(A->cja[A->cia[j]] == j);
#endif
	    CSR_NoDiag_colnorm2(A->ca[A->cia[j]], normptr);
	    for(k=A->cia[j]+1;k<A->cia[j+1];k++)
	      CSRcolnorm2(A->ca[k], normptr);
	  }
      else
	for(j=A->tlj;j<=A->brj;j++)
	  {
	    normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
#ifdef DEBUG_M
	    assert(A->cja[A->cia[j]] == j);
#endif
	    CSR_NoDiag_rownorm2(A->ca[A->cia[j]], normptr);
	    for(k=A->cia[j]+1;k<A->cia[j+1];k++)
	      CSRrownorm2(A->ca[k], normptr);
	  }
    }


  for(i=0;i<A->dim1;i++)
    normtab[i] = sqrt(normtab[i]);
}

void PhidalMatrix_ColNormINF(PhidalMatrix *A, PhidalHID *BL, REAL *normtab)
{
  dim_t i, j, k;
  REAL *normptr;

  bzero(normtab, sizeof(REAL)*A->dim2);
  if(A->csc == 0)
    for(j=A->tlj;j<=A->brj;j++)
      {
	normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	for(k=A->cia[j];k<A->cia[j+1];k++)
	  CSRcolnormINF(A->ca[k], normptr);
      }
  else
    for(j=A->tlj;j<=A->brj;j++)
      {
	normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	for(k=A->cia[j];k<A->cia[j+1];k++)
	  CSRrownormINF(A->ca[k], normptr);
      }

  if(A->symmetric == 1)
    {
      if(A->csc == 0)
	for(i=A->tli;i<=A->bri;i++)
	  {
	    normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	    for(k=A->ria[i];k<A->ria[i+1];k++)
	      CSRrownormINF(A->ra[k], normptr);
	  }
      else
	for(i=A->tli;i<=A->bri;i++)
	  {
	    normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	    for(k=A->ria[i];k<A->ria[i+1];k++)
	      CSRcolnormINF(A->ra[k], normptr);
	  }
    }


}

void PhidalMatrix_ColNorm1(PhidalMatrix *A, PhidalHID *BL, REAL *normtab)
{
  dim_t i, j, k;
  REAL *normptr;

  bzero(normtab, sizeof(REAL)*A->dim2);
  if(A->csc == 0)
    for(j=A->tlj;j<=A->brj;j++)
      {
	normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	for(k=A->cia[j];k<A->cia[j+1];k++)
	  CSRcolnorm1(A->ca[k], normptr);
      }
  else
    for(j=A->tlj;j<=A->brj;j++)
      {
	normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	for(k=A->cia[j];k<A->cia[j+1];k++)
	  CSRrownorm1(A->ca[k], normptr);
      }

  if(A->symmetric == 1)
    {
      if(A->csc == 0)
	for(i=A->tli;i<=A->bri;i++)
	  {
	    normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];

	    for(k=A->ria[i];k<A->ria[i+1]-1;k++)
	      CSRrownorm1(A->ra[k], normptr);
#ifdef DEBUG_M
	    assert(A->rja[A->ria[i+1]-1] == i);
#endif
	    CSR_NoDiag_rownorm1(A->ra[A->ria[i+1]-1], normptr);
	  }
      else
	for(i=A->tli;i<=A->bri;i++)
	  {
	    normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];

	    for(k=A->ria[i];k<A->ria[i+1]-1;k++)
	      CSRcolnorm1(A->ra[k], normptr);
#ifdef DEBUG_M
	    assert(A->rja[A->ria[i+1]-1] == i);
#endif
	    CSR_NoDiag_colnorm1(A->ra[A->ria[i+1]-1], normptr);
	  }
    }


}


void PhidalMatrix_ColNorm2(PhidalMatrix *A, PhidalHID *BL, REAL *normtab)
{
  dim_t i, j, k;
  REAL *normptr;

  bzero(normtab, sizeof(REAL)*A->dim2);
  if(A->csc == 0)
    for(j=A->tlj;j<=A->brj;j++)
      {
	normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	for(k=A->cia[j];k<A->cia[j+1];k++)
	  CSRcolnorm2(A->ca[k], normptr);
      }
  else
    for(j=A->tlj;j<=A->brj;j++)
      {
	normptr = normtab + BL->block_index[j] -BL->block_index[A->tlj];
	for(k=A->cia[j];k<A->cia[j+1];k++)
	  CSRrownorm2(A->ca[k], normptr);
      }

  
  if(A->symmetric == 1)
    {
      assert(A->csc == 1);
      if(A->csc == 0)
	for(i=A->tli;i<=A->bri;i++)
	  {
	    normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	    for(k=A->ria[i];k<A->ria[i+1]-1;k++)
	      CSRrownorm2(A->ra[k], normptr);
#ifdef DEBUG_M
	    assert(A->rja[A->ria[i+1]-1] == i);
#endif
	    CSR_NoDiag_rownorm2(A->ra[A->ria[i+1]-1], normptr); 
	  }
      else
	for(i=A->tli;i<=A->bri;i++)
	  {
	    normptr = normtab + BL->block_index[i] -BL->block_index[A->tli];
	    for(k=A->ria[i];k<A->ria[i+1]-1;k++)
	      CSRcolnorm2(A->ra[k], normptr);
#ifdef DEBUG_M
	    assert(A->rja[A->ria[i+1]-1] == i);
#endif
	    CSR_NoDiag_colnorm2(A->ra[A->ria[i+1]-1], normptr);
	  }
    }

  for(j=0;j<A->dim2;j++)
    normtab[j] = sqrt(normtab[j]);

}
