/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalMatrix_ReorderInterior.c                             */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>
/*#include "metis.h"*/

#include "phidal_sequential.h"




void PhidalMatrix_ReorderInterior(PhidalMatrix *A, PhidalHID *BL, dim_t *permtab, dim_t *ipermtab)
{
  /************************************************************************/
  /* This function reorder the matrices corresponding to interior         */
  /* in the Phidal Hierarchical decomposition                             */
  /* in order to minimize fill-in in the preconditionner                  */
  /* This function uses METIS 4.0                                         */
  /* Return:                                                              */
  /* the reordered matrix A                                               */
  /* permtab: the permutation vector permtab[i] is the new number of      */
  /* unknown i                                                            */
  /* ipermtab:  the inverse permutation vector                            */
  /*    i.e. ipermtab[i] is the former number of node i before the        */
  /* reordering                                                           */
  /************************************************************************/
  dim_t i;
  PhidalMatrix M;
  int start, end;
  int ii;

  start = BL->block_levelindex[0];
  end = BL->block_levelindex[1]-1;

  PhidalMatrix_BuildVirtualMatrix(start, start, end, end, A, &M, BL);
  /** Loop over all the interior matrices **/
  for(i=M.tli;i<=M.bri;i++)
    {
#ifdef DEBUG_M
      assert(M.ria[i+1]-M.ria[i] == 1);
#endif
      ii = BL->block_index[i]-BL->block_index[M.tli];
      CS_Reorder(M.ra[0], ipermtab+ii, permtab+ii);
    }
  
  /*** Leave the remaining of the node (connector of the interface) unchanged ***/
  for(i=BL->block_index[1];i<A->dim1;i++)
    {
      permtab[i] = i;
      ipermtab[i] = i;
    }

  PhidalMatrix_Clean(&M);
}




