/******************************************************************************/
/* Copyright 2008-2009 INRIA                                                  */
/*                                                                            */
/* File          : PhidalPrec_GEMM_ICCT.c                                     */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : P. HENON                                                   */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <math.h>

#include "phidal_sequential.h"


void PhidalPrec_GEMM_ICCT(int_t levelnum, REAL *droptab, PhidalMatrix *G, PhidalMatrix *A, COEF* DG, 
		      PhidalPrec *P, PhidalHID *BL, PhidalOptions *option)
{
  PhidalMatrix *VL;
  chrono_t t1, t2;
  COEF *D;

#ifdef DEBUG_M
  /*assert(option->schur_method == 2);*/
  assert(option->schur_method != 1);
  assert(A->tli == A->tlj);
  assert(A->bri == A->brj);
  assert(G->tli == A->tli);
  assert(G->bri == A->bri);
  assert(G->brj == A->tlj-1);
  assert(levelnum == option->forwardlev);
#endif

  t1 = dwalltime();
  P->dim = A->dim1;
  P->symmetric = 1;
  P->schur_method = option->schur_method;
  P->forwardlev = 0;
  P->levelnum =  levelnum;

  VL = (PhidalMatrix *)malloc(sizeof(PhidalMatrix));
  PhidalMatrix_Init(VL);
  PhidalMatrix_Setup(G->tlj, G->tlj, A->bri, A->bri, "L", "N", option->locally_nbr, VL, BL);

  PhidalMatrix_Cut(A, VL, BL);
  PhidalMatrix_Cut(G, VL, BL);
  
  D = (COEF *)malloc(sizeof(COEF)*VL->dim1);
  memcpy(D, DG, sizeof(COEF)*G->dim2);
  PhidalMatrix_ICCT_Restrict(A->tli, VL, D, option->droptol1, droptab, option->fillrat, BL, option, P->info);

  M_MALLOC(P->LU, SCAL);
  PREC_L_SCAL(P) = (PhidalMatrix *)malloc(sizeof(PhidalMatrix));
  PhidalMatrix_Setup(A->tli, A->tli, A->bri, A->bri, "L", "N", option->locally_nbr, PREC_L_SCAL(P), BL);
  PhidalMatrix_Cut(VL, PREC_L_SCAL(P), BL);

  P->D = (COEF *)malloc(sizeof(COEF)*A->dim1);
  assert(G->dim2+A->dim1 == VL->dim1);
  memcpy(P->D, D+G->dim2, sizeof(COEF)*A->dim1);
  free(D);
  PhidalMatrix_Clean(VL);
  free(VL);

  /*** Construct U as a virtual matrix equal to the transpose of L ***/
  PREC_U_SCAL(P) = (PhidalMatrix *)malloc(sizeof(PhidalMatrix));
  PhidalMatrix_Init(PREC_U_SCAL(P));
  PhidalMatrix_BuildVirtualMatrix(PREC_L_SCAL(P)->tli, PREC_L_SCAL(P)->tlj, PREC_L_SCAL(P)->bri, PREC_L_SCAL(P)->brj, PREC_L_SCAL(P), PREC_U_SCAL(P), BL); 
  PhidalMatrix_Transpose(PREC_U_SCAL(P));

  t2 = dwalltime();
  fprintfv(5, stderr, "Time in GEMM_ICCT = %g \n", t2-t1);
}
