
#undef TYPE_COMPLEX /* TMP FIX */







#ifdef INTSIZE32
#define INTS           INTEGER(KIND=4)
#define INTL           INTEGER(KIND=4)
#elif (defined INTSIZE64)
#define INTS           INTEGER(KIND=4)
#define INTL           INTEGER(KIND=8)
#elif (defined INTSSIZE64)
#define INTS           INTEGER(KIND=8) 
#define INTL           INTEGER(KIND=8) 
#else

#define INTS           INTEGER
#define INTL           INTEGER

#endif

#ifndef  PREC_SIMPLE

#define REALF REAL(KIND=8) 

#ifdef    TYPE_COMPLEX
#define COEF COMPLEX(KIND=8)
#else  
#define COEF REAL(KIND=8)
#endif 

#else  

#define REALF REAL(KIND=4)
#ifdef    TYPE_COMPLEX
#define COEF COMPLEX(KIND=4)
#else  
#define COEF REAL(KIND=4)
#endif 

/* real*4 real*8 */

#endif

!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
!     routines from SPARSKIT and extensions
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
      subroutine dperm (nrow,a,ja,ia,ao,jao,iao,perm,qperm,job)
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(nrow), &
              qperm(*),job
      COEF a(*),ao(*)
!-----------------------------------------------------------------------
! This routine permutes the rows and columns of a matrix stored in CSR
! format. i.e., it computes P A Q, where P, Q are permutation matrices.
! P maps row i into row perm(i) and Q maps column j into column qperm(j):
!      a(i,j)    becomes   a(perm(i),qperm(j)) in new matrix
! In the particular case where Q is the transpose of P (symmetric
! permutation of A) then qperm is not needed.
! note that qperm should be of length ncol (number of columns) but this
! is not checked.
!-----------------------------------------------------------------------
! Y. Saad, Sep. 21 1989 / recoded Jan. 28 1991.
!-----------------------------------------------------------------------
! on entry:
!----------
! n       = dimension of the matrix
! a, ja,
!    ia = input matrix in a, ja, ia format
! perm       = integer array of length n containing the permutation arrays
!        for the rows: perm(i) is the destination of row i in the
!         permuted matrix -- also the destination of column i in case
!         permutation is symmetric (job .le. 2)
!
! qperm      = same thing for the columns. This should be provided only
!         if job=3 or job=4, i.e., only in the case of a nonsymmetric
!        permutation of rows and columns. Otherwise qperm is a dummy
!
! job      = integer indicating the work to be done:
! * job = 1,2 permutation is symmetric  Ao :== P * A * transp(P)
!             job = 1      permute a, ja, ia into ao, jao, iao
!             job = 2 permute matrix ignoring real values.
! * job = 3,4 permutation is non-symmetric  Ao :== P * A * Q
!             job = 3      permute a, ja, ia into ao, jao, iao
!             job = 4 permute matrix ignoring real values.
!
! on return:
!-----------
! ao, jao, iao = input matrix in a, ja, ia format
!
! in case job .eq. 2 or job .eq. 4, a and ao are never referred to
! and can be dummy arguments.
! Notes:
!-------
!  1) algorithm is in place
!  2) column indices may not be sorted on return even  though they may be
!     on entry.
!----------------------------------------------------------------------c
! local variables
      integer locjob, mod
!
!     locjob indicates whether or not real values must be copied.
!
      locjob = mod(job,2)
!
! permute rows first
!
      call rperm (nrow,a,ja,ia,ao,jao,iao,perm,locjob)
!
! then permute columns
!
      locjob = 0
!
      if (job .le. 2) then
         call cperm (nrow,ao,jao,iao,ao,jao,iao,perm,locjob)
      else
         call cperm (nrow,ao,jao,iao,ao,jao,iao,qperm,locjob)
      endif
!
      return
!-------end-of-dperm----------------------------------------------------
!-----------------------------------------------------------------------
      end

!-----------------------------------------------------------------------
      subroutine dperm1 (i1,i2,a,ja,ia,b,jb,ib,perm,ipos,job)
      integer i1,i2,job,ja(*),ia(*),jb(*),ib(*),perm(*)
      COEF a(*),b(*)
!-----------------------------------------------------------------------
!     general submatrix permutation/ extraction routine.
!-----------------------------------------------------------------------
!     extracts rows perm(i1), perm(i1+1), ..., perm(i2) (in this order)
!     from a matrix (doing nothing in the column indices.) The resulting
!     submatrix is constructed in b, jb, ib. A pointer ipos to the
!     beginning of arrays b,jb,is also allowed (i.e., nonzero elements
!     are accumulated starting in position ipos of b, jb).
!-----------------------------------------------------------------------
! Y. Saad,Sep. 21 1989 / recoded Jan. 28 1991 / modified for SPARSLIB
! Sept. 1997..
!-----------------------------------------------------------------------
! on entry:
!----------
! n       = dimension of the matrix
! a,ja,
!   ia  = input matrix in CSR format
! perm       = integer array of length n containing the indices of the rows
!         to be extracted.
!
! job   = job indicator. if (job .ne.1) values are not copied (i.e.,
!         only pattern is copied).
!
! on return:
!-----------
! b,jb,
! ib   = matrix in csr format. b(ipos:ipos+nnz-1),jb(ipos:ipos+nnz-1)
!     contain the value and column indices respectively of the nnz
!     nonzero elements of the permuted matrix. thus ib(1)=ipos.
!
! Notes:
!-------
!  algorithm is NOT in place
!-----------------------------------------------------------------------
! local variables
!
      integer ko,irow,k
      logical values
!-----------------------------------------------------------------------
      values = (job .eq. 1)
      ko = ipos
      ib(1) = ko
      do 900 i=i1,i2
         irow = perm(i)
         do 800 k=ia(irow),ia(irow+1)-1
            if (values) b(ko) = a(k)
            jb(ko) = ja(k)
            ko=ko+1
 800     continue
         ib(i-i1+2) = ko
 900  continue
      return
!--------end-of-dperm1--------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine dperm2 (i1,i2,a,ja,ia,b,jb,ib,perm,rperm,istart, &
              ipos,job)
      integer i1,i2,job,istart,ja(*),ia(*),jb(*),ib(*),perm(*),rperm(*)
      COEF a(*),b(*)
!-----------------------------------------------------------------------
!     general submatrix permutation/ extraction routine.
!-----------------------------------------------------------------------
!     extracts rows rperm(i1), rperm(i1+1), ..., rperm(i2) and does an
!     associated column permutation (using array perm). The resulting
!     submatrix is constructed in b, jb, ib. For added flexibility,
!     the extracted are put in sequence starting from row 'istart' of B.
!     In addition a pointer ipos to the beginning of arrays b,jb,is also
!     allowed (i.e., nonzero elements are accumulated starting in
!     position ipos of b, jb).
!     extremely flexible. exple
!     to permute msr to msr (excluding diagonals)
!     call dperm2 (1,n,a,ja,ja,b,jb,jb,perm,rperm,1,n+2)
!-----------------------------------------------------------------------
! Y. Saad,Sep. 21 1989 / recoded Jan. 28 1991.
!-----------------------------------------------------------------------
! on entry:
!----------
! n       = dimension of the matrix
! a,ja,
!   ia  = input matrix in CSR format
! perm       = integer array of length n containing the permutation arrays
!        for the rows: perm(i) is the destination of row i in the
!         permuted matrix -- also the destination of column i.
!
! rperm      = reverse permutation. defined by rperm(iperm(i))=i for all i
!
! job   = job indicator. if (job .ne.1) values are not copied (i.e.,
!         only pattern is copied).
!
! on return:
!-----------
! b,ja,
! ib   = matrix in csr format. positions 1,2,...,istart-1 of ib
!     are not touched. b(ipos:ipos+nnz-1),jb(ipos:ipos+nnz-1)
!     contain the value and column indices respectively of the nnz
!     nonzero elements of the permuted matrix. thus ib(istart)=ipos.
!
! Notes:
!-------
!  1) algorithm is NOT in place
!  2) column indices may not be sorted on return even  though they
!     may be on entry.
!-----------------------------------------------------------------------
! local variables
!
      integer ko,irow,k
      logical values
!-----------------------------------------------------------------------
      values = (job .eq. 1)
      ko = ipos
      ib(istart) = ko
      do 900 i=i1,i2
         irow = rperm(i)
         do 800 k=ia(irow),ia(irow+1)-1
            if (values) b(ko) = a(k)
            jb(ko) = perm(ja(k))
            ko=ko+1
 800     continue
         ib(istart+i-i1+1) = ko
 900  continue
      return
!--------end-of-dperm2--------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine multic (n,ja,ia,ncol,kolrs,il,iord,maxcol,ierr)
      integer n, ja(*),ia(n+1),kolrs(n),iord(n),il(maxcol+1),ierr
!-----------------------------------------------------------------------
! multic:
!     multicoloring ordering -- greedy algorithm --
!     determines the coloring permutation and sets up
!     corresponding data structures for it.
!-----------------------------------------------------------------------
! on entry
! --------
! n     = row and column dimention of matrix
! ja    = column indices of nonzero elements of matrix, stored rowwise.
! ia    = pointer to beginning of each row in ja.
! maxcol= maximum number of colors allowed -- the size of il is
!         maxcol+1 at least. Note: the number of colors does not
!         exceed the maximum degree of each node +1.
!
! on return
! ---------
! ncol  = number of colours found
! kolrs = integer array containing the color number assigned to each node
! il    = integer array containing the pointers to the
!         beginning of each color set. In the permuted matrix
!         the rows /columns il(kol) to il(kol+1)-1 have the same color.
! iord  = permutation array corresponding to the multicolor ordering.
!         row number i will become row nbumber iord(i) in permuted
!         matrix. (iord = destination permutation array).
! ierr  = integer. Error message. normal return ierr = 0
!-----------------------------------------------------------------------
!
      integer kol, i, j, k, maxcol, mycol
!
      ierr = 0
      do 1 j=1, n
         kolrs(j) = 0
         iord(j) = j
 1    continue
      do 11 j=1, maxcol
         il(j) = 0
 11   continue
!
      ncol = 0
!
!     scan all nodes
!
      do 4 ii=1, n
         i = iord(ii)
!
!     look at adjacent nodes to determine colors already assigned
!
         mcol = 0
         do 2 k=ia(i), ia(i+1)-1
            j = ja(k)
            icol = kolrs(j)
            if (icol .ne. 0) then
               mcol = max(mcol,icol)
!
!     il used as temporary to record already assigned colors.
!
               il(icol) = 1
            endif
 2       continue
!
!     colors taken determined. scan il until a slot opens up.
!
         mycol = 1
 3       if (il(mycol) .eq. 1) then
            mycol = mycol+1
!
! #### put test elsewhere --
!
            if (mycol .gt. maxcol) goto 99
            if (mycol .le. mcol) goto 3
         endif
!
!     reset il to zero for next nodes
!
         do 35 j=1, mcol
            il(j) = 0
 35      continue
!
!     assign color and update number of colors so far
!
         kolrs(i) = mycol
         ncol = max(ncol,mycol)
 4    continue
!
!     every node has now been colored. Count nodes of each color
!
      do 6 j=1, n
         kol = kolrs(j)+1
         il(kol) = il(kol)+1
 6    continue
!
!     set pointers il
!
      il(1) = 1
      do 7 j=1, ncol
         il(j+1) = il(j)+il(j+1)
 7    continue
!
!     set iord
!
      do 8 j=1, n
         kol = kolrs(j)
         iord(j) = il(kol)
         il(kol) = il(kol)+1
 8    continue
!
!     shift il back
!
      do 9 j=ncol,1,-1
         il(j+1) = il(j)
 9    continue
      il(1) = 1
!
      return
 99   ierr = 1
      return
!-----end-of-multic-----------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine xtrows (i1,i2,a,ja,ia,ao,jao,iao,iperm,job)
      integer i1,i2,ja(*),ia(*),jao(*),iao(*),iperm(*),job
      COEF a(*),ao(*)
!-----------------------------------------------------------------------
! this subroutine extracts given rows from a matrix in CSR format.
! Specifically, rows number iperm(i1), iperm(i1+1), ...., iperm(i2)
! are extracted and put in the output matrix ao, jao, iao, in CSR
! format.  NOT in place.
! Youcef Saad -- coded Feb 15, 1992.
!-----------------------------------------------------------------------
! on entry:
!----------
! i1,i2   = two integers indicating the rows to be extracted.
!           xtrows will extract rows iperm(i1), iperm(i1+1),..,iperm(i2),
!           from original matrix and stack them in output matrix
!           ao, jao, iao in csr format
!
! a, ja, ia = input matrix in csr format
!
! iperm      = integer array of length nrow containing the reverse permutation
!         array for the rows. row number iperm(j) in permuted matrix PA
!         used to be row number j in unpermuted matrix.
!         ---> a(i,j) in the permuted matrix was a(iperm(i),j)
!         in the inout matrix.
!
! job      = integer indicating the work to be done:
!             job .ne. 1 : get structure only of output matrix,,
!               i.e., ignore real values. (in which case arrays a
!               and ao are not used nor accessed).
!             job = 1      get complete data structure of output matrix.
!               (i.e., including arrays ao and iao).
!------------
! on return:
!------------
! ao, jao, iao = input matrix in a, ja, ia format
! note :
!        if (job.ne.1)  then the arrays a and ao are not used.
!----------------------------------------------------------------------c
!           Y. Saad, revised May  2, 1990                              c
!----------------------------------------------------------------------c
      logical values
      values = (job .eq. 1)
!
! copying
!
      ko = 1
      iao(1) = ko
      do 100 j=i1,i2
!
! ii=iperm(j) is the index of old row to be copied.
!
         ii = iperm(j)
         do 60 k=ia(ii), ia(ii+1)-1
            jao(ko) = ja(k)
            if (values) ao(ko) = a(k)
            ko = ko+1
 60      continue
         iao(j-i1+2) = ko
 100  continue
!
      return
!---------end-of-xtrows-------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------

!-----------------------------------------------------------------------
      subroutine amub (nrow,ncol,job,a,ja,ia,b,jb,ib, &
                       c,jc,ic,nzmax,iw,ierr)
      COEF a(*), b(*), c(*)
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(ncol+1), &
           ic(ncol+1),iw(ncol)
!-----------------------------------------------------------------------
!     performs the matrix by matrix product C = A B
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow  = integer. The row dimension of A
! ncol  = integer. The column dimension of A
! job   = integer. Job indicator. When job = 0, only the structure
!                  (i.e. the arrays jc, ic) is computed and the
!                  real values are ignored.
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! b,
! jb,
! ib    =  Matrix B in compressed sparse row format.
!
! nzmax = integer. The  length of the arrays c and jc.
!         amub will stop if the result matrix C  has a number
!         of elements that exceeds exceeds nzmax. See ierr.
!
! on return:
!----------
! c,
! jc,
! ic    = resulting matrix C in compressed sparse row sparse format.
!
! ierr  = integer. serving as error message.
!         ierr = 0 means normal return,
!         ierr .gt. 0 means that amub stopped while computing the
!         i-th row  of C with i=ierr, because the number
!         of elements in C exceeds nzmax.
!
! work arrays:
!------------
! iw    = integer work array of length equal to the number of
!         columns in A.
! Notes:
!-------
!         The column dimension of B is not needed.
!
!-----------------------------------------------------------------
      COEF scal
      logical values
      values = (job .ne. 0)
      len = 0
      ic(1) = 1
      ierr = 0
!     initialize array iw.
      do 1 j=1, ncol
         iw(j) = 0
 1           continue
!
      do 500 ii=1, nrow
!     row i
         do 200 ka=ia(ii), ia(ii+1)-1
            if (values) scal = a(ka)
            jj   = ja(ka)
            do 100 kb=ib(jj),ib(jj+1)-1
               jcol = jb(kb)
               jpos = iw(jcol)
               if (jpos .eq. 0) then
                  len = len+1
                  if (len .gt. nzmax) then
                     ierr = ii
                     return
                  endif
                  jc(len) = jcol
                  iw(jcol)= len
                  if (values) c(len)  = scal*b(kb)
               else
                  if (values) c(jpos) = c(jpos) + scal*b(kb)
               endif
 100                   continue
 200                        continue
         do 201 k=ic(ii), len
            iw(jc(k)) = 0
 201             continue
         ic(ii+1) = len+1
 500       continue
      return
!-------------end-of-amub-----------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine dump (i1,i2,values,a,ja,ia,iout)
      integer i1, i2, ia(*), ja(*), iout
      COEF a(*)
      logical values
! outputs rows i1 through i2 of a sparse matrix stored in CSR format
! (or columns i1 through i2 of a matrix stored in CSC format) in a file,
! one (column) row at a time in a nice readable format.
! This is a simple routine which is useful for debugging.
!-----------------------------------------------------------------------
! on entry:
!---------
! i1    = first row (column) to print out
! i2    = last row (column) to print out
! values= logical. indicates whether or not to print real values.
!         if value = .false. only the pattern will be output.
! a,
! ja,
! ia    =  matrix in CSR format (or CSC format)
! iout  = logical unit number for output.
!----------
! the output file iout will have written in it the rows or columns
! of the matrix in one of two possible formats (depending on the max
! number of elements per row. The values are output with only
! two digits of accuracy (D9.2). )
!-----------------------------------------------------------------------
!     local variables
      integer maxr, i, k1, k2
!
! select mode horizontal or vertical
!
        maxr = 0
        do 1 i=i1, i2
           maxr = max0(maxr,ia(i+1)-ia(i))
 1               continue

        if (maxr .le. 8) then
!
! able to do one row acros line
!
        do 2 i=i1, i2
           write(iout,100) i
           k1=ia(i)
           k2 = ia(i+1)-1
           write (iout,101) (ja(k),k=k1,k2)
           if (values) write (iout,102) (a(k),k=k1,k2)
 2               continue
      else
!
! unable to one row acros line. do three items at a time
! across a line
         do 3 i=i1, i2
            if (values) then
               write(iout,200) i
            else
               write(iout,203) i
            endif
            k1=ia(i)
            k2 = ia(i+1)-1
            if (values) then
               write (iout,201) (ja(k),a(k),k=k1,k2)
            else
                write (iout,202) (ja(k),k=k1,k2)
             endif
 3                  continue
      endif
!
! formats :
!
 100    format (1h ,35(1h-),' row',i6,1x,35(1h-) )
 101      format(' col:',8(i5,6h     : ))
 102        format(' val:',8(D9.2,2h :) )
 200          format (1h ,31(1h-),' row',i3,1x,31(1h-),/ &
           3('  columns :    values  * ') )
!-------------xiiiiiihhhhhhddddddddd-*-
 201            format(3(1h ,i6,6h   :  ,D9.2,3h * ) )
 202              format(6(1h ,i5,6h  *    ) )
 203                format (1h ,31(1h-),' row',i3,1x,31(1h-),/ &
           3('  column  :  column   *') )
      return
!----end-of-dump--------------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine readmt (nmax,nzmax,job,iounit,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
!-----------------------------------------------------------------------
!     this subroutine reads  a boeing/harwell matrix. handles right hand
!     sides in full format only (no sparse right hand sides).
!     Also the matrix must be in assembled forms.
!     Author: Youcef Saad - Date: Sept., 1989
!     updated Oct 31, 1989.
!-----------------------------------------------------------------------
! on entry:
!---------
! nmax        =  max column dimension  allowed for matrix. The array ia should
!          be of length at least ncol+1 (see below) if job.gt.0
! nzmax       = max number of nonzeros elements allowed. the arrays a,
!          and ja should be of length equal to nnz (see below) if these
!          arrays are to be read (see job).
!
! job       = integer to indicate what is to be read. (note: job is an
!          input and output parameter, it can be modified on return)
!          job = 0    read the values of ncol, nrow, nnz, title, key,
!                     type and return. matrix is not read and arrays
!                     a, ja, ia, rhs are not touched.
!          job = 1    read srtucture only, i.e., the arrays ja and ia.
!          job = 2    read matrix including values, i.e., a, ja, ia
!          job = 3    read matrix and right hand sides: a,ja,ia,rhs.
!                  rhs may contain initial guesses and exact
!                     solutions appended to the actual right hand sides.
!                  this will be indicated by the output parameter
!                     guesol [see below].
!
! nrhs   = integer. nrhs is an input as well as ouput parameter.
!          at input nrhs contains the total length of the array rhs.
!          See also ierr and nrhs in output parameters.
!
! iounit = logical unit number where to read the matrix from.
!
! on return:
!----------
! job    = on return job may be modified to the highest job it could
!          do: if job=2 on entry but no matrix values are available it
!          is reset to job=1 on return. Similarly of job=3 but no rhs
!          is provided then it is rest to job=2 or job=1 depending on
!          whether or not matrix values are provided.
!          Note that no error message is triggered (i.e. ierr = 0
!          on return in these cases. It is therefore important to
!          compare the values of job on entry and return ).
!
! a       = the a matrix in the a, ia, ja (column) storage format
! ja        = row number of element a(i,j) in array a.
! ia     = pointer  array. ia(i) points to the beginning of column i.
!
! rhs    = real array of size nrow + 1 if available (see job)
!
! nrhs   = integer containing the number of right-hand sides found
!          each right hand side may be accompanied with an intial guess
!          and also the exact solution.
!
! guesol = a 2-character string indicating whether an initial guess
!          (1-st character) and / or the exact solution (2-nd
!          character) is provided with the right hand side.
!         if the first character of guesol is 'G' it means that an
!          an intial guess is provided for each right-hand side.
!          These are appended to the right hand-sides in the array rhs.
!         if the second character of guesol is 'X' it means that an
!          exact solution is provided for each right-hand side.
!          These are  appended to the right hand-sides
!          and the initial guesses (if any) in the array rhs.
!
! nrow   = number of rows in matrix
! ncol       = number of columns in matrix
! nnz       = number of nonzero elements in A. This info is returned
!          even if there is not enough space in a, ja, ia, in order
!          to determine the minimum storage needed.
!
! title  = character*100 = title of matrix test ( character a*72).
! key    = character*8  = key of matrix
! type   = charatcer*3  = type of matrix.
!          for meaning of title, key and type refer to documentation
!          Harwell/Boeing matrices.
!
! ierr   = integer used for error messages
!         * ierr  =  0 means that  the matrix has been read normally.
!         * ierr  =  1 means that  the array matrix could not be read
!         because ncol+1 .gt. nmax
!         * ierr  =  2 means that  the array matrix could not be read
!         because nnz .gt. nzmax
!         * ierr  =  3 means that  the array matrix could not be read
!         because both (ncol+1 .gt. nmax) and  (nnz .gt. nzmax )
!         * ierr  =  4 means that  the right hand side (s) initial
!         guesse (s) and exact solution (s)   could  not be
!         read because they are stored in sparse format (not handled
!         by this routine ...)
!         * ierr  =  5 means that the right-hand-sides, initial guesses
!         and exact solutions could not be read because the length of
!         rhs as specified by the input value of nrhs is not
!         sufficient to store them. The rest of the matrix may have
!         been read normally.
!
! Notes:
!-------
! 1) The file inout must be open (and possibly rewound if necessary)
!    prior to calling readmt.
! 2) Refer to the documentation on the Harwell-Boeing formats
!    for details on the format assumed by readmt.
!    We summarize the format here for convenience.
!
!    a) all lines in inout are assumed to be 80 character long.
!    b) the file consists of a header followed by the block of the
!       column start pointers followed by the block of the
!       row indices, followed by the block of the real values and
!       finally the numerical values of the right-hand-side if a
!       right hand side is supplied.
!    c) the file starts by a header which contains four lines if no
!       right hand side is supplied and five lines otherwise.
!       * first line contains the title (72 characters long) followed by
!         the 8-character identifier (name of the matrix, called key)
!        [ A72,A8 ]
!       * second line contains the number of lines for each
!         of the following data blocks (4 of them) and the total number
!         of lines excluding the header.
!        [5i4]
!       * the third line contains a three character string identifying
!         the type of matrices as they are referenced in the Harwell
!         Boeing documentation [e.g., rua, rsa,..] and the number of
!         rows, columns, nonzero entries.
!         [A3,11X,4I14]
!       * The fourth line contains the variable fortran format
!         for the following data blocks.
!         [2A16,2A20]
!       * The fifth line is only present if right-hand-sides are
!         supplied. It consists of three one character-strings containing
!         the storage format for the right-hand-sides
!         ('F'= full,'M'=sparse=same as matrix), an initial guess
!         indicator ('G' for yes), an exact solution indicator
!         ('X' for yes), followed by the number of right-hand-sides
!         and then the number of row indices.
!         [A3,11X,2I14]
!     d) The three following blocks follow the header as described
!        above.
!     e) In case the right hand-side are in sparse formats then
!        the fourth block uses the same storage format as for the matrix
!        to describe the NRHS right hand sides provided, with a column
!        being replaced by a right hand side.
!-----------------------------------------------------------------------
        character title*72, key*8, type*3, ptrfmt*16, indfmt*16, &
           valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol, &
           nnz, neltvl, nrhs, nmax, nzmax, nrwindx
      integer ia (nmax+1), ja (nzmax)
      COEF a(nzmax), rhs(*)
!-----------------------------------------------------------------------
      ierr = 0
      lenrhs = nrhs
!
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd, &
           rhscrd, type, nrow, ncol, nnz, neltvl, ptrfmt, indfmt, &
           valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
!
      if (rhscrd .gt. 0) read (iounit,11) rhstyp, nrhs, nrwindx
 11   format (a3,11x,i14,i14)
!
!     anything else to read ?
!
      if (job .le. 0) return
!     ---- check whether matrix is readable ------
      n = ncol
      if (ncol .gt. nmax) ierr = 1
      if (nnz .gt. nzmax) ierr = ierr + 2
      if (ierr .ne. 0) return
!     ---- read pointer and row numbers ----------
      read (iounit,ptrfmt) (ia (i), i = 1, n+1)
      read (iounit,indfmt) (ja (i), i = 1, nnz)
!     --- reading values of matrix if required....
      if (job .le. 1)  return
!     --- and if available -----------------------
      if (valcrd .le. 0) then
       job = 1
       return
      endif
      read (iounit,valfmt) (a(i), i = 1, nnz)
!     --- reading rhs if required ----------------
      if (job .le. 2)  return
!     --- and if available -----------------------
      if ( rhscrd .le. 0) then
       job = 2
       return
      endif
!
!     --- read right-hand-side.--------------------
!
      if (rhstyp(1:1) .eq. 'M') then
         ierr = 4
         return
      endif
!
      guesol = rhstyp(2:3)
!
      nvec = 1
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') nvec=nvec+1
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') nvec=nvec+1
!
      len = nrhs*nrow
!
      if (len*nvec .gt. lenrhs) then
         ierr = 5
         return
      endif
!
!     read right-hand-sides
!
      next = 1
      iend = len
      read(iounit,rhsfmt) (rhs(i), i = next, iend)
!
!     read initial guesses if available
!
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
         next = next+len
         iend = iend+ len
         read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
!     read exact solutions if available
!
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
         next = next+len
         iend = iend+ len
         read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
      return
!---------end-of-readmt-------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine roscal(nrow,job,nrm,a,ja,ia,diag,b,jb,ib,ierr)
      COEF a(*), b(*), diag(nrow)
      integer nrow,job,nrm,ja(*),jb(*),ia(nrow+1),ib(nrow+1),ierr
!-----------------------------------------------------------------------
! scales the rows of A such that their norms are one on return
! 3 choices of norms: 1-norm, 2-norm, max-norm.
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A
!
! job   = integer. job indicator. Job=0 means get array b only
!         job = 1 means get b, and the integer arrays ib, jb.
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! on return:
!----------
!
! diag = diagonal matrix stored as a vector containing the matrix
!        by which the rows have been scaled, i.e., on return
!        we have B = Diag*A.
!
! b,
! jb,
! ib      = resulting matrix B in compressed sparse row sparse format.
!
! ierr  = error message. ierr=0     : Normal return
!                        ierr=i > 0 : Row number i is a zero row.
! Notes:
!-------
! 1)        The column dimension of A is not needed.
! 2)        algorithm in place (B can take the place of A).
!-----------------------------------------------------------------
      call rnrms (nrow,nrm,a,ja,ia,diag)

      ierr = 0
      do 1 j=1, nrow
         if (diag(j) .eq. 0.0d0) then
            ierr = j
            return
         else
            diag(j) = 1.0d0/diag(j)
         endif
 1    continue
      call diamua(nrow,job,a,ja,ia,diag,b,jb,ib)
      return
!-------end-of-roscal---------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine coscal(nrow,job,nrm,a,ja,ia,diag,b,jb,ib,ierr)
!-----------------------------------------------------------------------
      COEF a(*),b(*),diag(nrow)
      integer nrow,job,ja(*),jb(*),ia(nrow+1),ib(nrow+1),ierr
!-----------------------------------------------------------------------
! scales the columns of A such that their norms are one on return
! result matrix written on b, or overwritten on A.
! 3 choices of norms: 1-norm, 2-norm, max-norm. in place.
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A
!
! job   = integer. job indicator. Job=0 means get array b only
!         job = 1 means get b, and the integer arrays ib, jb.
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! on return:
!----------
!
! diag = diagonal matrix stored as a vector containing the matrix
!        by which the columns have been scaled, i.e., on return
!        we have B = A * Diag
!
! b,
! jb,
! ib      = resulting matrix B in compressed sparse row sparse format.
!
! ierr  = error message. ierr=0     : Normal return
!                        ierr=i > 0 : Column number i is a zero row.
! Notes:
!-------
! 1)        The column dimension of A is not needed.
! 2)       algorithm in place (B can take the place of A).
!-----------------------------------------------------------------
      call cnrms (nrow,nrm,a,ja,ia,diag)
      ierr = 0
      do 1 j=1, nrow
         if (diag(j) .eq. 0.0) then
            ierr = j
            return
         else
            diag(j) = 1.0d0/diag(j)
         endif
 1    continue
      call amudia (nrow,job,a,ja,ia,diag,b,jb,ib)
      return
!--------end-of-coscal--------------------------------------------------
!-----------------------------------------------------------------------
      end
      subroutine rnrms   (nrow, nrm, a, ja, ia, diag)
      COEF a(*), diag(nrow), scal
      integer ja(*), ia(nrow+1)
!-----------------------------------------------------------------------
! gets the norms of each row of A. (choice of three norms)
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! on return:
!----------
!
! diag = real vector of length nrow containing the norms
!
!-----------------------------------------------------------------
      do 1 ii=1,nrow
!
!     compute the norm if each element.
!
         scal = 0.0d0
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         if (nrm .eq. 0) then
            do 2 k=k1, k2
               scal = max(scal,abs(a(k) ) )
 2          continue
         elseif (nrm .eq. 1) then
            do 3 k=k1, k2
               scal = scal + abs(a(k) )
 3          continue
         else
            do 4 k=k1, k2
               scal = scal+a(k)**2
 4          continue
         endif
         if (nrm .eq. 2) scal = sqrt(scal)
         diag(ii) = scal
 1    continue
      return
!-----------------------------------------------------------------------
!-------------end-of-rnrms----------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine cnrms   (nrow, nrm, a, ja, ia, diag)
      COEF a(*), diag(nrow)
      integer ja(*), ia(nrow+1)
!-----------------------------------------------------------------------
! gets the norms of each column of A. (choice of three norms)
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A
!
! nrm   = integer. norm indicator. nrm = 1, means 1-norm, nrm =2
!                  means the 2-nrm, nrm = 0 means max norm
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! on return:
!----------
!
! diag = real vector of length nrow containing the norms
!
!-----------------------------------------------------------------
      do 10 k=1, nrow
         diag(k) = 0.0d0
 10   continue
      do 1 ii=1,nrow
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         do 2 k=k1, k2
            j = ja(k)
!     update the norm of each column
            if (nrm .eq. 0) then
               diag(j) = max(diag(j),abs(a(k) ) )
            elseif (nrm .eq. 1) then
               diag(j) = diag(j) + abs(a(k) )
            else
               diag(j) = diag(j)+a(k)**2
            endif
 2       continue
 1    continue
      if (nrm .ne. 2) return
      do 3 k=1, nrow
         diag(k) = sqrt(diag(k))
 3    continue
      return
!-----------------------------------------------------------------------
!------------end-of-cnrms-----------------------------------------------
      end
      subroutine diamua (nrow,job, a, ja, ia, diag, b, jb, ib)
      COEF a(*), b(*), diag(nrow), scal
      integer ja(*),jb(*), ia(nrow+1),ib(nrow+1)
!-----------------------------------------------------------------------
! performs the matrix by matrix product B = Diag * A  (in place)
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A
!
! job   = integer. job indicator. Job=0 means get array b only
!         job = 1 means get b, and the integer arrays ib, jb.
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! diag = diagonal matrix stored as a vector dig(1:n)
!
! on return:
!----------
!
! b,
! jb,
! ib      = resulting matrix B in compressed sparse row sparse format.
!
! Notes:
!-------
! 1)        The column dimension of A is not needed.
! 2)        algorithm in place (B can take the place of A).
!           in this case use job=0.
!-----------------------------------------------------------------
      do 1 ii=1,nrow
!
!     normalize each row
!
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         scal = diag(ii)
         do 2 k=k1, k2
            b(k) = a(k)*scal
 2       continue
 1    continue
!
      if (job .eq. 0) return
!
      do 3 ii=1, nrow+1
         ib(ii) = ia(ii)
 3    continue
      do 31 k=ia(1), ia(nrow+1) -1
         jb(k) = ja(k)
 31   continue
      return
!----------end-of-diamua------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine amudia (nrow,job, a, ja, ia, diag, b, jb, ib)
      COEF a(*), b(*), diag(nrow)
      integer ja(*),jb(*), ia(nrow+1),ib(nrow+1)
!-----------------------------------------------------------------------
! performs the matrix by matrix product B = A * Diag  (in place)
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A
!
! job   = integer. job indicator. Job=0 means get array b only
!         job = 1 means get b, and the integer arrays ib, jb.
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! diag = diagonal matrix stored as a vector dig(1:n)
!
! on return:
!----------
!
! b,
! jb,
! ib      = resulting matrix B in compressed sparse row sparse format.
!
! Notes:
!-------
! 1)        The column dimension of A is not needed.
! 2)        algorithm in place (B can take the place of A).
!-----------------------------------------------------------------
      do 1 ii=1,nrow
!
!     scale each element
!
         k1 = ia(ii)
         k2 = ia(ii+1)-1
         do 2 k=k1, k2
            b(k) = a(k)*diag(ja(k))
 2       continue
 1    continue
!
      if (job .eq. 0) return
!
      do 3 ii=1, nrow+1
         ib(ii) = ia(ii)
 3    continue
      do 31 k=ia(1), ia(nrow+1) -1
         jb(k) = ja(k)
 31   continue
      return
!-----------------------------------------------------------------------
!-----------end-of-amudiag----------------------------------------------
      end
! aplb   :   computes     C = A+B                                      c
      subroutine faplb (nrow,ncol,job,a,ja,ia,b,jb,ib, &
           c,jc,ic,nzmax,iw,ierr)
      COEF a(*), b(*), c(*)
      integer ja(*),jb(*),jc(*),ia(nrow+1),ib(nrow+1),ic(nrow+1), &
           iw(ncol)
!-----------------------------------------------------------------------
! performs the matrix sum  C = A+B.
!-----------------------------------------------------------------------
! on entry:
! ---------
! nrow      = integer. The row dimension of A and B
! ncol  = integer. The column dimension of A and B.
! job   = integer. Job indicator. When job = 0, only the structure
!                  (i.e. the arrays jc, ic) is computed and the
!                  real values are ignored.
!
! a,
! ja,
! ia   = Matrix A in compressed sparse row format.
!
! b,
! jb,
! ib      =  Matrix B in compressed sparse row format.
!
! nzmax      = integer. The  length of the arrays c and jc.
!         amub will stop if the result matrix C  has a number
!         of elements that exceeds exceeds nzmax. See ierr.
!
! on return:
!----------
! c,
! jc,
! ic      = resulting matrix C in compressed sparse row sparse format.
!
! ierr      = integer. serving as error message.
!         ierr = 0 means normal return,
!         ierr .gt. 0 means that amub stopped while computing the
!         i-th row  of C with i=ierr, because the number
!         of elements in C exceeds nzmax.
!
! work arrays:
!------------
! iw      = integer work array of length equal to the number of
!         columns in A.
!
!-----------------------------------------------------------------------
      logical values

      values = (job .ne. 0)
      ierr = 0
      len = 0
      ic(1) = 1
      do 1 j=1, ncol
         iw(j) = 0
 1    continue
!
      do 500 ii=1, nrow
!     row i
         do 200 ka=ia(ii), ia(ii+1)-1
            len = len+1
            jcol    = ja(ka)
            if (len .gt. nzmax) goto 999
            jc(len) = jcol
            if (values) c(len)  = a(ka)
            iw(jcol)= len
 200     continue
!
         do 300 kb=ib(ii),ib(ii+1)-1
            jcol = jb(kb)
            jpos = iw(jcol)
            if (jpos .eq. 0) then
               len = len+1
               if (len .gt. nzmax) goto 999
               jc(len) = jcol
               if (values) c(len)  = b(kb)
               iw(jcol)= len
            else
               if (values) c(jpos) = c(jpos) + b(kb)
            endif
 300     continue
         do 301 k=ic(ii), len
          iw(jc(k)) = 0
 301     continue
         ic(ii+1) = len+1
 500  continue
      return
 999  ierr = ii
      return
!------------end of aplb -----------------------------------------------
      end
! csrcsc  : converts compressed sparse row format to compressed sparse c
      subroutine fcsrcsc (n,job,ipos,a,ja,ia,ao,jao,iao)
      integer ia(n+1),iao(n+1),ja(*),jao(*)
      COEF  a(*),ao(*)
!-----------------------------------------------------------------------
! Compressed Sparse Row     to      Compressed Sparse Column
!
! (transposition operation)   Not in place.
!-----------------------------------------------------------------------
! -- not in place --
! this subroutine transposes a matrix stored in a, ja, ia format.
! ---------------
! on entry:
!----------
! n      = dimension of A.
! job      = integer to indicate whether to fill the values (job.eq.1) of the
!         matrix ao or only the pattern., i.e.,ia, and ja (job .ne.1)
!
! ipos  = starting position in ao, jao of the transposed matrix.
!         the iao array takes this into account (thus iao(1) is set to ipos.)
!         Note: this may be useful if one needs to append the data structure
!         of the transpose to that of A. In this case use for example
!                call csrcsc (n,1,n+2,a,ja,ia,a,ja,ia(n+2))
!        for any other normal usage, enter ipos=1.
! a      = real array of length nnz (nnz=number of nonzero elements in input
!         matrix) containing the nonzero elements.
! ja      = integer array of length nnz containing the column positions
!         of the corresponding elements in a.
! ia      = integer of size n+1. ia(k) contains the position in a, ja of
!        the beginning of the k-th row.
!
! on return:
! ----------
! output arguments:
! ao      = real array of size nzz containing the "a" part of the transpose
! jao      = integer array of size nnz containing the column indices.
! iao      = integer array of size n+1 containing the "ia" index array of
!        the transpose.
!
!-----------------------------------------------------------------------
      call fcsrcsc2 (n,n,job,ipos,a,ja,ia,ao,jao,iao)
      end
!-----------------------------------------------------------------------
      subroutine fcsrcsc2 (n,n2,job,ipos,a,ja,ia,ao,jao,iao)
      integer ia(n+1),iao(n2+1),ja(*),jao(*)
      COEF  a(*),ao(*)
!-----------------------------------------------------------------------
! Compressed Sparse Row     to      Compressed Sparse Column
!
! (transposition operation)   Not in place.
!-----------------------------------------------------------------------
! Rectangular version.  n is number of rows of CSR matrix,
!                       n2 (input) is number of columns of CSC matrix.
!-----------------------------------------------------------------------
! -- not in place --
! this subroutine transposes a matrix stored in a, ja, ia format.
! ---------------
! on entry:
!----------
! n      = number of rows of CSR matrix.
! n2    = number of columns of CSC matrix.
! job      = integer to indicate whether to fill the values (job.eq.1) of the
!         matrix ao or only the pattern., i.e.,ia, and ja (job .ne.1)
!
! ipos  = starting position in ao, jao of the transposed matrix.
!         the iao array takes this into account (thus iao(1) is set to ipos.)
!         Note: this may be useful if one needs to append the data structure
!         of the transpose to that of A. In this case use for example
!                call csrcsc2 (n,n,1,n+2,a,ja,ia,a,ja,ia(n+2))
!        for any other normal usage, enter ipos=1.
! a      = real array of length nnz (nnz=number of nonzero elements in input
!         matrix) containing the nonzero elements.
! ja      = integer array of length nnz containing the column positions
!         of the corresponding elements in a.
! ia      = integer of size n+1. ia(k) contains the position in a, ja of
!        the beginning of the k-th row.
!
! on return:
! ----------
! output arguments:
! ao      = real array of size nzz containing the "a" part of the transpose
! jao      = integer array of size nnz containing the column indices.
! iao      = integer array of size n+1 containing the "ia" index array of
!        the transpose.
!
!-----------------------------------------------------------------------
!----------------- compute lengths of rows of transp(A) ----------------
      do 1 i=1,n2+1
         iao(i) = 0
 1    continue
      do 3 i=1, n
         do 2 k=ia(i), ia(i+1)-1
            j = ja(k)+1
            iao(j) = iao(j)+1
 2       continue
 3    continue
!---------- compute pointers from lengths ------------------------------
      iao(1) = ipos
      do 4 i=1,n2
         iao(i+1) = iao(i) + iao(i+1)
 4    continue
!--------------- now do the actual copying -----------------------------
      do 6 i=1,n
         do 62 k=ia(i),ia(i+1)-1
            j = ja(k)
            next = iao(j)
            if (job .eq. 1)  ao(next) = a(k)
            jao(next) = i
            iao(j) = next+1
 62      continue
 6    continue
!-------------------------- reshift iao and leave ----------------------
      do 7 i=n2,1,-1
         iao(i+1) = iao(i)
 7    continue
      iao(1) = ipos
!--------------- end of csrcsc2 ----------------------------------------
!-----------------------------------------------------------------------
      end

!-----------------------------------------------------------------------
      subroutine atmux (n, x, y, a, ja, ia)
      COEF x(*), y(*), a(*)
      integer n, ia(*), ja(*)
!-----------------------------------------------------------------------
!         transp( A ) times a vector
!-----------------------------------------------------------------------
! multiplies the transpose of a matrix by a vector when the original
! matrix is stored in compressed sparse row storage. Can also be
! viewed as the product of a matrix by a vector when the original
! matrix is stored in the compressed sparse column format.
!-----------------------------------------------------------------------
!
! on entry:
!----------
! n     = row dimension of A
! x     = real array of length equal to the column dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
!-----------
! y     = real array of length n, containing the product y=transp(A)*x
!
!-----------------------------------------------------------------------
!     local variables
!
      integer i, k
!-----------------------------------------------------------------------
!
!     zero out output vector
!
      do 1 i=1,n
         y(i) = 0.0
 1    continue
!
! loop over the rows
!
      do 100 i = 1,n
         do 99 k=ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
 99      continue
 100  continue
!
      return
!-------------end-of-atmux----------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine atmuxr (m, n, x, y, a, ja, ia)
      COEF x(*), y(*), a(*)
      integer m, n, ia(*), ja(*)
!-----------------------------------------------------------------------
!         transp( A ) times a vector, A can be rectangular
!-----------------------------------------------------------------------
! See also atmux.  The essential difference is how the solution vector
! is initially zeroed.  If using this to multiply rectangular CSC
! matrices by a vector, m number of rows, n is number of columns.
!-----------------------------------------------------------------------
!
! on entry:
!----------
! m     = column dimension of A
! n     = row dimension of A
! x     = real array of length equal to the column dimension of
!         the A matrix.
! a, ja,
!    ia = input matrix in compressed sparse row format.
!
! on return:
!-----------
! y     = real array of length n, containing the product y=transp(A)*x
!
!-----------------------------------------------------------------------
!     local variables
!
      integer i, k
!-----------------------------------------------------------------------
!
!     zero out output vector
!
      do 1 i=1,m
         y(i) = 0.0
 1    continue
!
! loop over the rows
!
      do 100 i = 1,n
         do 99 k=ia(i), ia(i+1)-1
            y(ja(k)) = y(ja(k)) + x(i)*a(k)
 99      continue
 100  continue
!
      return
!-------------end-of-atmuxr---------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine readmt_c (nmax,nzmax,job,fname,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
!-----------------------------------------------------------------------
!     this  subroutine reads  a boeing/harwell matrix,  given the
!     corresponding file. handles right hand sides in full format
!     only (no sparse right hand sides). Also the matrix must  be
!     in assembled forms.
!     It differs from readmt, in that  the name of the file needs
!     to be passed, and then the file is opened and closed within
!     this routine.
!     Author: Youcef Saad - Date: Oct 31, 1989
!     updated Jul 20, 1998 by Irene Moulitsas
!-----------------------------------------------------------------------
! on entry:
!---------
! nmax        =  max column dimension  allowed for matrix. The array ia should
!          be of length at least ncol+1 (see below) if job.gt.0
! nzmax       = max number of nonzeros elements allowed. the arrays a,
!          and ja should be of length equal to nnz (see below) if these
!          arrays are to be read (see job).
!
! job       = integer to indicate what is to be read. (note: job is an
!          input and output parameter, it can be modified on return)
!          job = 0    read the values of ncol, nrow, nnz, title, key,
!                     type and return. matrix is not read and arrays
!                     a, ja, ia, rhs are not touched.
!          job = 1    read srtucture only, i.e., the arrays ja and ia.
!          job = 2    read matrix including values, i.e., a, ja, ia
!          job = 3    read matrix and right hand sides: a,ja,ia,rhs.
!                  rhs may contain initial guesses and exact
!                     solutions appended to the actual right hand sides.
!                  this will be indicated by the output parameter
!                     guesol [see below].
!
! fname = name of the file where to read the matrix from.
!
! nrhs   = integer. nrhs is an input as well as ouput parameter.
!          at input nrhs contains the total length of the array rhs.
!          See also ierr and nrhs in output parameters.
!
! on return:
!----------
! job    = on return job may be modified to the highest job it could
!          do: if job=2 on entry but no matrix values are available it
!          is reset to job=1 on return. Similarly of job=3 but no rhs
!          is provided then it is rest to job=2 or job=1 depending on
!          whether or not matrix values are provided.
!          Note that no error message is triggered (i.e. ierr = 0
!          on return in these cases. It is therefore important to
!          compare the values of job on entry and return ).
!
! a       = the a matrix in the a, ia, ja (column) storage format
! ja        = column number of element a(i,j) in array a.
! ia     = pointer  array. ia(i) points to the beginning of column i.
!
! rhs    = real array of size nrow + 1 if available (see job)
!
! nrhs   = integer containing the number of right-hand sides found
!          each right hand side may be accompanied with an intial guess
!          and also the exact solution.
!
! guesol = a 2-character string indicating whether an initial guess
!          (1-st character) and / or the exact solution (2-nd
!          character) is provided with the right hand side.
!         if the first character of guesol is 'G' it means that an
!          an intial guess is provided for each right-hand side.
!          These are appended to the right hand-sides in the array rhs.
!         if the second character of guesol is 'X' it means that an
!          exact solution is provided for each right-hand side.
!          These are  appended to the right hand-sides
!          and the initial guesses (if any) in the array rhs.
!
! nrow   = number of rows in matrix
! ncol       = number of columns in matrix
! nnz       = number of nonzero elements in A. This info is returned
!          even if there is not enough space in a, ja, ia, in order
!          to determine the minimum storage needed.
!
! title  = character*72 = title of matrix test ( character a*72).
! key    = character*8  = key of matrix
! type   = charatcer*3  = type of matrix.
!          for meaning of title, key and type refer to documentation
!          Harwell/Boeing matrices.
!
! ierr   = integer used for error messages
!         * ierr  =  0 means that  the matrix has been read normally.
!         * ierr  =  1 means that  the array matrix could not be read
!         because ncol+1 .gt. nmax
!         * ierr  =  2 means that  the array matrix could not be read
!         because nnz .gt. nzmax
!         * ierr  =  3 means that  the array matrix could not be read
!         because both (ncol+1 .gt. nmax) and  (nnz .gt. nzmax )
!         * ierr  =  4 means that  the right hand side (s) initial
!         guesse (s) and exact solution (s)   could  not be
!         read because they are stored in sparse format (not handled
!         by this routine ...)
!         * ierr  =  5 means that the right-hand-sides, initial guesses
!         and exact solutions could not be read because the length of
!         rhs as specified by the input value of nrhs is not
!         sufficient to store them. The rest of the matrix may have
!         been read normally.
!
! Notes:
!-------
! 1) This routine can be interfaced with the C language, since only
!    the name of the  file needs to be passed and no iounti number.
!
! 2) Refer to the  documentation on  the Harwell-Boeing formats for
!    details on the format assumed by readmt.
!    We summarize the format here for convenience.
!
!    a) all  lines in  inout are  assumed to be 80  character long.
!    b) the file  consists of a header followed by the block of the
!       column start  pointers followed  by  the  block  of the row
!       indices,  followed  by  the  block  of the  real values and
!       finally the  numerical  values of  the right-hand-side if a
!       right hand side is supplied.
!    c) the file starts by a header which contains four lines if no
!       right hand side is supplied and five lines otherwise.
!       * first  line  contains  the  title  (72  characters  long)
!         followed  by  the  8-character  identifier (name  of  the
!         matrix, called key) [ A72,A8 ]
!       * second line  contains the number of lines for each of the
!         following data blocks (4 of them) and the total number of
!         lines excluding the header.  [5i4]
!       * the   third  line  contains  a  three   character  string
!         identifying the type of  matrices as they  are referenced
!         in  the Harwell  Boeing documentation [e.g., rua, rsa,..]
!         and the number of rows, columns, nonzero entries.
!         [A3,11X,4I14]
!       * The fourth  line contains the variable fortran format for
!         the following data blocks. [2A16,2A20]
!       * The fifth  line is  only present if  right-hand-sides are
!         supplied. It  consists  of  three  one  character-strings
!         containing the  storage  format for the  right-hand-sides
!         ('F'= full,'M'=sparse=same as matrix), an  initial  guess
!         indicator  ('G' for yes),  an  exact  solution  indicator
!         ('X' for yes), followed by the number of right-hand-sides
!         and then the number of row indices.  [A3,11X,2I14]
!     d) The three  following blocks follow the header as described
!        above.
!     e) In case the right hand-side are in sparse formats then the
!        fourth  block  uses the  same  storage  format as  for the
!        matrix to  describe  the NRHS right  hand  sides provided,
!        with a column being replaced by a right hand side.
!-----------------------------------------------------------------------
      character title*72, key*8, type*3, ptrfmt*16, indfmt*16, &
           valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol, &
           nnz, neltvl, nrhs, nmax, nzmax, nrwindx
      integer ia (nmax+1), ja (nzmax)
      COEF a(nzmax), rhs(*)
      character fname*100
!-----------------------------------------------------------------------
      ierr = 0
      lenrhs = nrhs
!
      iounit=15
      open(iounit,file = fname)
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd, &
           rhscrd, type, nrow, ncol, nnz, neltvl, ptrfmt, indfmt, &
           valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
!
      if (rhscrd .gt. 0) read (iounit,11) rhstyp, nrhs, nrwindx
 11   format (a3,11x,i14,i14)
!
!     anything else to read ?
!
      if (job .le. 0) goto 12

!     ---- check whether matrix is readable ------
      n = ncol
      if (ncol .gt. nmax) ierr = 1
      if (nnz .gt. nzmax) ierr = ierr + 2
      if (ierr .ne. 0) goto 12

!     ---- read pointer and row numbers ----------
      read (iounit,ptrfmt) (ia (i), i = 1, n+1)
      read (iounit,indfmt) (ja (i), i = 1, nnz)

!     --- reading values of matrix if required....
      if (job .le. 1)  goto 12
!     --- and if available -----------------------
      if (valcrd .le. 0) then
       job = 1
       goto 12
      endif
      read (iounit,valfmt) (a(i), i = 1, nnz)

!     --- reading rhs if required ----------------
      if (job .le. 2)  goto 12
!     --- and if available -----------------------
      if ( rhscrd .le. 0) then
       job = 2
       goto 12
      endif
!
!     --- read right-hand-side.--------------------
!
      if (rhstyp(1:1) .eq. 'M') then
         ierr = 4
         goto 12
      endif
!
      guesol = rhstyp(2:3)
!
      nvec = 1
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') nvec=nvec+1
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') nvec=nvec+1
!
      len = nrhs*nrow
!
      if (len*nvec .gt. lenrhs) then
        ierr = 5
        goto 12
      endif
!
!     read right-hand-sides
!
      next = 1
      iend = len
      read(iounit,rhsfmt) (rhs(i), i = next, iend)
!
!     read initial guesses if available
!
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
        next = next+len
        iend = iend+ len
        read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
!     read exact solutions if available
!
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
        next = next+len
        iend = iend+ len
        read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
 12   close(iounit)
      return
!---------end-of-readmt_c-------------------------------------------------
!-------------------------------------------------------------------------
      end

!-------------------------------------------------------------------------
      subroutine csort (n,a,ja,ia,iwork,values)
      logical values
      integer n, ja(*), ia(n+1), iwork(*)
      COEF a(*)
!-----------------------------------------------------------------------
! This routine sorts the elements of  a matrix (stored in Compressed
! Sparse Row Format) in increasing order of their column indices within
! each row. It uses a form of bucket sort with a cost of O(nnz) where
! nnz = number of nonzero elements.
! requires an integer work array of length 2*nnz.
!-----------------------------------------------------------------------
! on entry:
!---------
! n     = the row dimension of the matrix
! a     = the matrix A in compressed sparse row format.
! ja    = the array of column indices of the elements in array a.
! ia    = the array of pointers to the rows.
! iwork = integer work array of length max ( n+1, 2*nnz )
!         where nnz = (ia(n+1)-ia(1))  ) .
! values= logical indicating whether or not the real values a(*) must
!         also be permuted. if (.not. values) then the array a is not
!         touched by csort and can be a dummy array.
!
! on return:
!----------
! the matrix stored in the structure a, ja, ia is permuted in such a
! way that the column indices are in increasing order within each row.
! iwork(1:nnz) contains the permutation used  to rearrange the elements.
!-----------------------------------------------------------------------
! Y. Saad - Feb. 1, 1991.
! revised by Zhongze Li - June 13th, 2001
!-----------------------------------------------------------------------
! local variables
      integer i, k, j, ifirst, offset, nnz, next
!
! count the number of elements in each column
!
      offset = 1 - ia(1)
      do 1 i=1,n+1
         iwork(i) = 0
 1    continue
      do 3 i=1, n
         do 2 k=ia(i), ia(i+1)-1
            j = ja(k+offset)+1+offset
            iwork(j) = iwork(j)+1
 2       continue
 3    continue
!
! compute pointers from lengths.
!
      iwork(1) = 1
      do 4 i=1,n
         iwork(i+1) = iwork(i) + iwork(i+1)
 4    continue
!
! get the positions of the nonzero elements in order of columns.
!
      ifirst = ia(1);
      nnz = ia(n+1)-ifirst
      do 5 i=1,n
         do 51 k=ia(i),ia(i+1)-1
            j = ja(k+offset)+offset
            next = iwork(j)
            iwork(nnz+next) = k+offset
            iwork(j) = next+1
 51      continue
 5    continue
!
! convert to coordinate format
!
      do 6 i=1, n
         do 61 k=ia(i), ia(i+1)-1
            iwork(k+offset) = i
 61      continue
 6    continue
!
! loop to find permutation: for each element find the correct
! position in (sorted) arrays a, ja. Record this in iwork.
!
      do 7 k=1, nnz
         ko = iwork(nnz+k)
         irow = iwork(ko)
         next = ia(irow)
!
! the current element should go in next position in row. iwork
! records this position.
!
         iwork(ko) = next+offset
         ia(irow)  = next+1
 7       continue
!
! perform an in-place permutation of the  arrays.
!
         call ivperm (nnz, ja, iwork)
         if (values) call dvperm (nnz, a, iwork)
!
! reshift the pointers of the original matrix back.
!
      do 8 i=n,1,-1
         ia(i+1) = ia(i)
 8    continue
      ia(1) = ifirst
!
      return
!---------------end-of-csort--------------------------------------------
!-----------------------------------------------------------------------
      end

!-----------------------------------------------------------------------
      subroutine dvperm (n, x, perm)
      integer n, perm(n)
      COEF x(n)
!-----------------------------------------------------------------------
! this subroutine performs an in-place permutation of a real vector x
! according to the permutation array perm(*), i.e., on return,
! the vector x satisfies,
!
!      x(perm(j)) :== x(j), j=1,2,.., n
!
!-----------------------------------------------------------------------
! on entry:
!---------
! n       = length of vector x.
! perm       = integer array of length n containing the permutation  array.
! x      = input vector
!
! on return:
!----------
! x      = vector x permuted according to x(perm(*)) :=  x(*)
!
!----------------------------------------------------------------------c
!           Y. Saad, Sep. 21 1989                                      c
!----------------------------------------------------------------------c
! local variables
      COEF tmp, tmp1
!
      init      = 1
      tmp      = x(init)
      ii        = perm(init)
      perm(init)= -perm(init)
      k         = 0
!
! loop
!
 6    k = k+1
!
! save the chased element --
!
      tmp1        = x(ii)
      x(ii)     = tmp
      next        = perm(ii)
      if (next .lt. 0 ) goto 65
!
! test for end
!
      if (k .gt. n) goto 101
      tmp       = tmp1
      perm(ii)  = - perm(ii)
      ii        = next
!
! end loop
!
      goto 6
!
! reinitilaize cycle --
!
 65   init      = init+1
      if (init .gt. n) goto 101
      if (perm(init) .lt. 0) goto 65
      tmp      = x(init)
      ii      = perm(init)
      perm(init)=-perm(init)
      goto 6
!
 101  continue
      do 200 j=1, n
         perm(j) = -perm(j)
 200  continue
!
      return
!-------------------end-of-dvperm---------------------------------------
!-----------------------------------------------------------------------
      end

!-----------------------------------------------------------------------
      subroutine ivperm (n, ix, perm)
      integer n, perm(n), ix(n)
!-----------------------------------------------------------------------
! this subroutine performs an in-place permutation of an integer vector
! ix according to the permutation array perm(*), i.e., on return,
! the vector x satisfies,
!
!      ix(perm(j)) :== ix(j), j=1,2,.., n
!
!-----------------------------------------------------------------------
! on entry:
!---------
! n       = length of vector x.
! perm       = integer array of length n containing the permutation  array.
! ix      = input vector
!
! on return:
!----------
! ix      = vector x permuted according to ix(perm(*)) :=  ix(*)
!
!----------------------------------------------------------------------c
!           Y. Saad, Sep. 21 1989                                      c
!----------------------------------------------------------------------c
! local variables
      integer tmp, tmp1
!
      init      = 1
      tmp      = ix(init)
      ii        = perm(init)
      perm(init)= -perm(init)
      k         = 0
!
! loop
!
 6    k = k+1
!
! save the chased element --
!
      tmp1        = ix(ii)
      ix(ii)     = tmp
      next        = perm(ii)
      if (next .lt. 0 ) goto 65
!
! test for end
!
      if (k .gt. n) goto 101
      tmp       = tmp1
      perm(ii)  = - perm(ii)
      ii        = next
!
! end loop
!
      goto 6
!
! reinitilaize cycle --
!
 65   init      = init+1
      if (init .gt. n) goto 101
      if (perm(init) .lt. 0) goto 65
      tmp      = ix(init)
      ii      = perm(init)
      perm(init)=-perm(init)
      goto 6
!
 101  continue
      do 200 j=1, n
         perm(j) = -perm(j)
 200  continue
!
      return
!-------------------end-of-ivperm---------------------------------------
!-----------------------------------------------------------------------
      end


!-----------------------------------------------------------------------
      subroutine rperm (nrow,a,ja,ia,ao,jao,iao,perm,job)
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(nrow),job
      COEF a(*),ao(*)
!-----------------------------------------------------------------------
! this subroutine permutes the rows of a matrix in CSR format.
! rperm  computes B = P A  where P is a permutation matrix.
! the permutation P is defined through the array perm: for each j,
! perm(j) represents the destination row number of row number j.
! Youcef Saad -- recoded Jan 28, 1991.
!-----------------------------------------------------------------------
! on entry:
!----------
! n       = dimension of the matrix
! a, ja, ia = input matrix in csr format
! perm       = integer array of length nrow containing the permutation arrays
!        for the rows: perm(i) is the destination of row i in the
!         permuted matrix.
!         ---> a(i,j) in the original matrix becomes a(perm(i),j)
!         in the output  matrix.
!
! job      = integer indicating the work to be done:
!             job = 1      permute a, ja, ia into ao, jao, iao
!                       (including the copying of real values ao and
!                       the array iao).
!             job .ne. 1 :  ignore real values.
!                     (in which case arrays a and ao are not needed nor
!                      used).
!
!------------
! on return:
!------------
! ao, jao, iao = input matrix in a, ja, ia format
! note :
!        if (job.ne.1)  then the arrays a and ao are not used.
!----------------------------------------------------------------------c
!           Y. Saad, May  2, 1990                                      c
!----------------------------------------------------------------------c
      logical values
      values = (job .eq. 1)
!
!     determine pointers for output matix.
!
      do 50 j=1,nrow
         i = perm(j)
         iao(i+1) = ia(j+1) - ia(j)
 50   continue
!
! get pointers from lengths
!
      iao(1) = 1
      do 51 j=1,nrow
         iao(j+1)=iao(j+1)+iao(j)
 51   continue
!
! copying
!
      do 100 ii=1,nrow
!
! old row = ii  -- new row = iperm(ii) -- ko = new pointer
!
         ko = iao(perm(ii))
         do 60 k=ia(ii), ia(ii+1)-1
            jao(ko) = ja(k)
            if (values) ao(ko) = a(k)
            ko = ko+1
 60      continue
 100  continue
!
      return
!---------end-of-rperm -------------------------------------------------
!-----------------------------------------------------------------------
      end

      subroutine cperm (nrow,a,ja,ia,ao,jao,iao,perm,job)
      integer nrow,ja(*),ia(nrow+1),jao(*),iao(nrow+1),perm(*), job
      COEF a(*), ao(*)
!-----------------------------------------------------------------------
! this subroutine permutes the columns of a matrix a, ja, ia.
! the result is written in the output matrix  ao, jao, iao.
! cperm computes B = A P, where  P is a permutation matrix
! that maps column j into column perm(j), i.e., on return
!      a(i,j) becomes a(i,perm(j)) in new matrix
! Y. Saad, May 2, 1990 / modified Jan. 28, 1991.
!-----------------------------------------------------------------------
! on entry:
!----------
! nrow       = row dimension of the matrix
!
! a, ja, ia = input matrix in csr format.
!
! perm      = integer array of length ncol (number of columns of A
!         containing the permutation array  the columns:
!         a(i,j) in the original matrix becomes a(i,perm(j))
!         in the output matrix.
!
! job      = integer indicating the work to be done:
!             job = 1      permute a, ja, ia into ao, jao, iao
!                       (including the copying of real values ao and
!                       the array iao).
!             job .ne. 1 :  ignore real values ao and ignore iao.
!
!------------
! on return:
!------------
! ao, jao, iao = input matrix in a, ja, ia format (array ao not needed)
!
! Notes:
!-------
! 1. if job=1 then ao, iao are not used.
! 2. This routine is in place: ja, jao can be the same.
! 3. If the matrix is initially sorted (by increasing column number)
!    then ao,jao,iao  may not be on return.
!
!----------------------------------------------------------------------c
! local parameters:
      integer k, i, nnz
!
      nnz = ia(nrow+1)-1
      do 100 k=1,nnz
         jao(k) = perm(ja(k))
 100  continue
!
!     done with ja array. return if no need to touch values.
!
      if (job .ne. 1) return
!
! else get new pointers -- and copy values too.
!
      do 1 i=1, nrow+1
         iao(i) = ia(i)
 1    continue
!
      do 2 k=1, nnz
         ao(k) = a(k)
 2    continue
!
      return
!---------end-of-cperm--------------------------------------------------
!-----------------------------------------------------------------------
      end
!
!
!
      subroutine wreadmtc (nmax,nzmax,job,fname,length,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
!-----------------------------------------------------------------------
!     this  subroutine reads  a boeing/harwell matrix,  given the
!     corresponding file. handles right hand sides in full format
!     only (no sparse right hand sides). Also the matrix must  be
!     in assembled forms.
!     It differs from readmt, in that  the name of the file needs
!     to be passed, and then the file is opened and closed within
!     this routine.
!     Author: Youcef Saad - Date: Oct 31, 1989
!     updated Feb 6th, 2001 by Zhongze Li
!-----------------------------------------------------------------------
      character fname*100, title*72, key*8, type*3, guesol*2, fname1*100
      integer nrow, ncol, nnz, nrhs, nmax, nzmax, job, length, ierr
      integer ia (nmax+1), ja (nzmax)
      COEF a(nzmax), rhs(*)

      fname1 = " "
      fname1(1:length) = fname(1:length)
      call readmtc(nmax,nzmax,job,fname1,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
      return
      end

      subroutine readmtc (nmax,nzmax,job,fname,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
!-----------------------------------------------------------------------
!     this  subroutine reads  a boeing/harwell matrix,  given the
!     corresponding file. handles right hand sides in full format
!     only (no sparse right hand sides). Also the matrix must  be
!     in assembled forms.
!     It differs from readmt, in that  the name of the file needs
!     to be passed, and then the file is opened and closed within
!     this routine.
!     Author: Youcef Saad - Date: Oct 31, 1989
!     updated Jul 20, 1998 by Irene Moulitsas
!-----------------------------------------------------------------------
! on entry:
!---------
! nmax   =  max column dimension  allowed for matrix. The array ia should
!           be of length at least ncol+1 (see below) if job.gt.0
! nzmax  = max number of nonzeros elements allowed. the arrays a,
!          and ja should be of length equal to nnz (see below) if these
!          arrays are to be read (see job).
!
! job    = integer to indicate what is to be read. (note: job is an
!          input and output parameter, it can be modified on return)
!          job = 0    read the values of ncol, nrow, nnz, title, key,
!                     type and return. matrix is not read and arrays
!                     a, ja, ia, rhs are not touched.
!          job = 1    read srtucture only, i.e., the arrays ja and ia.
!          job = 2    read matrix including values, i.e., a, ja, ia
!          job = 3    read matrix and right hand sides: a,ja,ia,rhs.
!                     rhs may contain initial guesses and exact
!                     solutions appended to the actual right hand sides.
!                     this will be indicated by the output parameter
!                     guesol [see below].
!
! fname = name of the file where to read the matrix from.
!
! nrhs   = integer. nrhs is an input as well as ouput parameter.
!          at input nrhs contains the total length of the array rhs.
!          See also ierr and nrhs in output parameters.
!
! on return:
!----------
! job    = on return job may be modified to the highest job it could
!          do: if job=2 on entry but no matrix values are available it
!          is reset to job=1 on return. Similarly of job=3 but no rhs
!          is provided then it is rest to job=2 or job=1 depending on
!          whether or not matrix values are provided.
!          Note that no error message is triggered (i.e. ierr = 0
!          on return in these cases. It is therefore important to
!          compare the values of job on entry and return ).
!
! a      = the a matrix in the a, ia, ja (column) storage format
! ja     = column number of element a(i,j) in array a.
! ia     = pointer  array. ia(i) points to the beginning of column i.
!
! rhs    = real array of size nrow + 1 if available (see job)
!
! nrhs   = integer containing the number of right-hand sides found
!          each right hand side may be accompanied with an intial guess
!          and also the exact solution.
!
! guesol = a 2-character string indicating whether an initial guess
!          (1-st character) and / or the exact solution (2-nd
!          character) is provided with the right hand side.
!          if the first character of guesol is 'G' it means that an
!          an intial guess is provided for each right-hand side.
!          These are appended to the right hand-sides in the array rhs.
!          if the second character of guesol is 'X' it means that an
!          exact solution is provided for each right-hand side.
!          These are  appended to the right hand-sides
!          and the initial guesses (if any) in the array rhs.
!
! nrow   = number of rows in matrix
! ncol   = number of columns in matrix
! nnz    = number of nonzero elements in A. This info is returned
!          even if there is not enough space in a, ja, ia, in order
!          to determine the minimum storage needed.
!
! title  = character*72 = title of matrix test ( character a*72).
! key    = character*8  = key of matrix
! type   = charatcer*3  = type of matrix.
!          for meaning of title, key and type refer to documentation
!          Harwell/Boeing matrices.
!
! ierr   = integer used for error messages
!         * ierr  =  0 means that  the matrix has been read normally.
!         * ierr  =  1 means that  the array matrix could not be read
!         because ncol+1 .gt. nmax
!         * ierr  =  2 means that  the array matrix could not be read
!         because nnz .gt. nzmax
!         * ierr  =  3 means that  the array matrix could not be read
!         because both (ncol+1 .gt. nmax) and  (nnz .gt. nzmax )
!         * ierr  =  4 means that  the right hand side (s) initial
!         guesse (s) and exact solution (s)   could  not be
!         read because they are stored in sparse format (not handled
!         by this routine ...)
!         * ierr  =  5 means that the right-hand-sides, initial guesses
!         and exact solutions could not be read because the length of
!         rhs as specified by the input value of nrhs is not
!         sufficient to store them. The rest of the matrix may have
!         been read normally.
!
! Notes:
!-------
! 1) This routine can be interfaced with the C language, since only
!    the name of the  file needs to be passed and no iounti number.
!
! 2) Refer to the  documentation on  the Harwell-Boeing formats for
!    details on the format assumed by readmt.
!    We summarize the format here for convenience.
!
!    a) all  lines in  inout are  assumed to be 80  character long.
!    b) the file  consists of a header followed by the block of the
!       column start  pointers followed  by  the  block  of the row
!       indices,  followed  by  the  block  of the  real values and
!       finally the  numerical  values of  the right-hand-side if a
!       right hand side is supplied.
!    c) the file starts by a header which contains four lines if no
!       right hand side is supplied and five lines otherwise.
!       * first  line  contains  the  title  (72  characters  long)
!         followed  by  the  8-character  identifier (name  of  the
!         matrix, called key) [ A72,A8 ]
!       * second line  contains the number of lines for each of the
!         following data blocks (4 of them) and the total number of
!         lines excluding the header.  [5i4]
!       * the   third  line  contains  a  three   character  string
!         identifying the type of  matrices as they  are referenced
!         in  the Harwell  Boeing documentation [e.g., rua, rsa,..]
!         and the number of rows, columns, nonzero entries.
!         [A3,11X,4I14]
!       * The fourth  line contains the variable fortran format for
!         the following data blocks. [2A16,2A20]
!       * The fifth  line is  only present if  right-hand-sides are
!         supplied. It  consists  of  three  one  character-strings
!         containing the  storage  format for the  right-hand-sides
!         ('F'= full,'M'=sparse=same as matrix), an  initial  guess
!         indicator  ('G' for yes),  an  exact  solution  indicator
!         ('X' for yes), followed by the number of right-hand-sides
!         and then the number of row indices.  [A3,11X,2I14]
!     d) The three  following blocks follow the header as described
!        above.
!     e) In case the right hand-side are in sparse formats then the
!        fourth  block  uses the  same  storage  format as  for the
!        matrix to  describe  the NRHS right  hand  sides provided,
!        with a column being replaced by a right hand side.
!-----------------------------------------------------------------------
      character title*72, key*8, type*3, ptrfmt*16, indfmt*16, &
           valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol, &
           nnz, neltvl, nrhs, nmax, nzmax, nrwindx
      integer ia (nmax+1), ja (nzmax)
      COEF a(nzmax), rhs(*)
      character fname*100
!-----------------------------------------------------------------------
      ierr = 0
      lenrhs = nrhs
!
      iounit=15
      open(iounit,file = fname)
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd, &
           rhscrd, type, nrow, ncol, nnz, neltvl, ptrfmt, indfmt, &
           valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
!
      if (rhscrd .gt. 0) read (iounit,11) rhstyp, nrhs, nrwindx
 11   format (a3,11x,i14,i14)
!
!     anything else to read ?
!
      if (job .le. 0) goto 12
!     ---- check whether matrix is readable ------
      n = ncol
      if (ncol .gt. nmax) ierr = 1
      if (nnz .gt. nzmax) ierr = ierr + 2
      if (ierr .ne. 0) goto 12
!     ---- read pointer and row numbers ----------
      read (iounit,ptrfmt) (ia (i), i = 1, n+1)
      read (iounit,indfmt) (ja (i), i = 1, nnz)
!     --- reading values of matrix if required....
      if (job .le. 1)  goto 12
!     --- and if available -----------------------
      if (valcrd .le. 0) then
         job = 1
         goto 12
      endif
      read (iounit,valfmt) (a(i), i = 1, nnz)
!     --- reading rhs if required ----------------
      if (job .le. 2)  goto 12
!     --- and if available -----------------------
      if ( rhscrd .le. 0) then
         job = 2
         goto 12
      endif
!
!     --- read right-hand-side.--------------------
!
      if (rhstyp(1:1) .eq. 'M') then
         ierr = 4
         goto 12
      endif
!
      guesol = rhstyp(2:3)
!
      nvec = 1
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') nvec=nvec+1
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') nvec=nvec+1
!
      len = nrhs*nrow
!
      if (len*nvec .gt. lenrhs) then
        ierr = 5
        goto 12
      endif
!
!     read right-hand-sides
!
      next = 1
      iend = len
      read(iounit,rhsfmt) (rhs(i), i = next, iend)
!
!     read initial guesses if available
!
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
        next = next+len
        iend = iend+ len
        read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
!     read exact solutions if available
!
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
        next = next+len
        iend = iend+ len
        read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
 12   close(iounit)
      return
!---------end-of-readmt_c-------------------------------------------------
!-----------------------------------------------------------------------
      end

!-----------------------------------------------------------------------
!------------------ 64 bits wreadmtc ---- modified P. HENON ------------
!-----------------------------------------------------------------------

      subroutine wreadmtc64 (nmax,nzmax,job,length,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr, fname)
!-----------------------------------------------------------------------
!     this  subroutine reads  a boeing/harwell matrix,  given the
!     corresponding file. handles right hand sides in full format
!     only (no sparse right hand sides). Also the matrix must  be
!     in assembled forms.
!     It differs from readmt, in that  the name of the file needs
!     to be passed, and then the file is opened and closed within
!     this routine.
!     Author: Youcef Saad - Date: Oct 31, 1989
!     updated Feb 6th, 2001 by Zhongze Li
!-----------------------------------------------------------------------
      character fname*100, title*72, key*8, type*3, guesol*2, fname1*100
      integer*8 nrow, ncol, nnz, nrhs, nmax, nzmax, job, length, ierr
      integer*8 ia (nmax+1), ja (nzmax)
      COEF a(nzmax), rhs(*)

      fname1 = " "

!      print *,'length ',length
!      print *,'fname ',fname
!      print *,'fname1 ',fname1

      fname1(1:length) = fname(1:length)
      call readmtc64(nmax,nzmax,job,fname1,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
      return
      end

      subroutine readmtc64 (nmax,nzmax,job,fname,a,ja,ia,rhs,nrhs, &
           guesol,nrow,ncol,nnz,title,key,type,ierr)
!-----------------------------------------------------------------------
!     this  subroutine reads  a boeing/harwell matrix,  given the
!     corresponding file. handles right hand sides in full format
!     only (no sparse right hand sides). Also the matrix must  be
!     in assembled forms.
!     It differs from readmt, in that  the name of the file needs
!     to be passed, and then the file is opened and closed within
!     this routine.
!     Author: Youcef Saad - Date: Oct 31, 1989
!     updated Jul 20, 1998 by Irene Moulitsas
!-----------------------------------------------------------------------
! on entry:
!---------
! nmax   =  max column dimension  allowed for matrix. The array ia should
!           be of length at least ncol+1 (see below) if job.gt.0
! nzmax  = max number of nonzeros elements allowed. the arrays a,
!          and ja should be of length equal to nnz (see below) if these
!          arrays are to be read (see job).
!
! job    = integer to indicate what is to be read. (note: job is an
!          input and output parameter, it can be modified on return)
!          job = 0    read the values of ncol, nrow, nnz, title, key,
!                     type and return. matrix is not read and arrays
!                     a, ja, ia, rhs are not touched.
!          job = 1    read srtucture only, i.e., the arrays ja and ia.
!          job = 2    read matrix including values, i.e., a, ja, ia
!          job = 3    read matrix and right hand sides: a,ja,ia,rhs.
!                     rhs may contain initial guesses and exact
!                     solutions appended to the actual right hand sides.
!                     this will be indicated by the output parameter
!                     guesol [see below].
!
! fname = name of the file where to read the matrix from.
!
! nrhs   = integer. nrhs is an input as well as ouput parameter.
!          at input nrhs contains the total length of the array rhs.
!          See also ierr and nrhs in output parameters.
!
! on return:
!----------
! job    = on return job may be modified to the highest job it could
!          do: if job=2 on entry but no matrix values are available it
!          is reset to job=1 on return. Similarly of job=3 but no rhs
!          is provided then it is rest to job=2 or job=1 depending on
!          whether or not matrix values are provided.
!          Note that no error message is triggered (i.e. ierr = 0
!          on return in these cases. It is therefore important to
!          compare the values of job on entry and return ).
!
! a      = the a matrix in the a, ia, ja (column) storage format
! ja     = column number of element a(i,j) in array a.
! ia     = pointer  array. ia(i) points to the beginning of column i.
!
! rhs    = real array of size nrow + 1 if available (see job)
!
! nrhs   = integer containing the number of right-hand sides found
!          each right hand side may be accompanied with an intial guess
!          and also the exact solution.
!
! guesol = a 2-character string indicating whether an initial guess
!          (1-st character) and / or the exact solution (2-nd
!          character) is provided with the right hand side.
!          if the first character of guesol is 'G' it means that an
!          an intial guess is provided for each right-hand side.
!          These are appended to the right hand-sides in the array rhs.
!          if the second character of guesol is 'X' it means that an
!          exact solution is provided for each right-hand side.
!          These are  appended to the right hand-sides
!          and the initial guesses (if any) in the array rhs.
!
! nrow   = number of rows in matrix
! ncol   = number of columns in matrix
! nnz    = number of nonzero elements in A. This info is returned
!          even if there is not enough space in a, ja, ia, in order
!          to determine the minimum storage needed.
!
! title  = character*72 = title of matrix test ( character a*72).
! key    = character*8  = key of matrix
! type   = charatcer*3  = type of matrix.
!          for meaning of title, key and type refer to documentation
!          Harwell/Boeing matrices.
!
! ierr   = integer used for error messages
!         * ierr  =  0 means that  the matrix has been read normally.
!         * ierr  =  1 means that  the array matrix could not be read
!         because ncol+1 .gt. nmax
!         * ierr  =  2 means that  the array matrix could not be read
!         because nnz .gt. nzmax
!         * ierr  =  3 means that  the array matrix could not be read
!         because both (ncol+1 .gt. nmax) and  (nnz .gt. nzmax )
!         * ierr  =  4 means that  the right hand side (s) initial
!         guesse (s) and exact solution (s)   could  not be
!         read because they are stored in sparse format (not handled
!         by this routine ...)
!         * ierr  =  5 means that the right-hand-sides, initial guesses
!         and exact solutions could not be read because the length of
!         rhs as specified by the input value of nrhs is not
!         sufficient to store them. The rest of the matrix may have
!         been read normally.
!
! Notes:
!-------
! 1) This routine can be interfaced with the C language, since only
!    the name of the  file needs to be passed and no iounti number.
!
! 2) Refer to the  documentation on  the Harwell-Boeing formats for
!    details on the format assumed by readmt.
!    We summarize the format here for convenience.
!
!    a) all  lines in  inout are  assumed to be 80  character long.
!    b) the file  consists of a header followed by the block of the
!       column start  pointers followed  by  the  block  of the row
!       indices,  followed  by  the  block  of the  real values and
!       finally the  numerical  values of  the right-hand-side if a
!       right hand side is supplied.
!    c) the file starts by a header which contains four lines if no
!       right hand side is supplied and five lines otherwise.
!       * first  line  contains  the  title  (72  characters  long)
!         followed  by  the  8-character  identifier (name  of  the
!         matrix, called key) [ A72,A8 ]
!       * second line  contains the number of lines for each of the
!         following data blocks (4 of them) and the total number of
!         lines excluding the header.  [5i4]
!       * the   third  line  contains  a  three   character  string
!         identifying the type of  matrices as they  are referenced
!         in  the Harwell  Boeing documentation [e.g., rua, rsa,..]
!         and the number of rows, columns, nonzero entries.
!         [A3,11X,4I14]
!       * The fourth  line contains the variable fortran format for
!         the following data blocks. [2A16,2A20]
!       * The fifth  line is  only present if  right-hand-sides are
!         supplied. It  consists  of  three  one  character-strings
!         containing the  storage  format for the  right-hand-sides
!         ('F'= full,'M'=sparse=same as matrix), an  initial  guess
!         indicator  ('G' for yes),  an  exact  solution  indicator
!         ('X' for yes), followed by the number of right-hand-sides
!         and then the number of row indices.  [A3,11X,2I14]
!     d) The three  following blocks follow the header as described
!        above.
!     e) In case the right hand-side are in sparse formats then the
!        fourth  block  uses the  same  storage  format as  for the
!        matrix to  describe  the NRHS right  hand  sides provided,
!        with a column being replaced by a right hand side.
!-----------------------------------------------------------------------
      character title*72, key*8, type*3, ptrfmt*16, indfmt*16, &
           valfmt*20, rhsfmt*20, rhstyp*3, guesol*2
      integer*8 totcrd, ptrcrd, indcrd, valcrd, rhscrd, nrow, ncol, &
           nnz, neltvl, nrhs, nmax, nzmax, nrwindx, job, ierr
      integer*8 ia (nmax+1), ja (nzmax)
      COEF a(nzmax), rhs(*)
      character fname*100
!-----------------------------------------------------------------------
      ierr = 0
      lenrhs = nrhs
!
      iounit=15
      open(iounit,file = fname)
      read (iounit,10) title, key, totcrd, ptrcrd, indcrd, valcrd, &
           rhscrd, type, nrow, ncol, nnz, neltvl, ptrfmt, indfmt, &
           valfmt, rhsfmt
 10   format (a72, a8 / 5i14 / a3, 11x, 4i14 / 2a16, 2a20)
!
      if (rhscrd .gt. 0) read (iounit,11) rhstyp, nrhs, nrwindx
 11   format (a3,11x,i14,i14)
!
!     anything else to read ?
!
      if (job .le. 0) goto 12
!     ---- check whether matrix is readable ------
      n = ncol
      if (ncol .gt. nmax) ierr = 1
      if (nnz .gt. nzmax) ierr = ierr + 2
      if (ierr .ne. 0) goto 12
!     ---- read pointer and row numbers ----------
      read (iounit,ptrfmt) (ia (i), i = 1, n+1)
      read (iounit,indfmt) (ja (i), i = 1, nnz)
!     --- reading values of matrix if required....
      if (job .le. 1)  goto 12
!     --- and if available -----------------------
      if (valcrd .le. 0) then
         job = 1
         goto 12
      endif
      read (iounit,valfmt) (a(i), i = 1, nnz)
!     --- reading rhs if required ----------------
      if (job .le. 2)  goto 12
!     --- and if available -----------------------
      if ( rhscrd .le. 0) then
         job = 2
         goto 12
      endif
!
!     --- read right-hand-side.--------------------
!
      if (rhstyp(1:1) .eq. 'M') then
         ierr = 4
         goto 12
      endif
!
      guesol = rhstyp(2:3)
!
      nvec = 1
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') nvec=nvec+1
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') nvec=nvec+1
!
      len = nrhs*nrow
!
      if (len*nvec .gt. lenrhs) then
        ierr = 5
        goto 12
      endif
!
!     read right-hand-sides
!
      next = 1
      iend = len
      read(iounit,rhsfmt) (rhs(i), i = next, iend)
!
!     read initial guesses if available
!
      if (guesol(1:1) .eq. 'G' .or. guesol(1:1) .eq. 'g') then
        next = next+len
        iend = iend+ len
        read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
!     read exact solutions if available
!
      if (guesol(2:2) .eq. 'X' .or. guesol(2:2) .eq. 'x') then
        next = next+len
        iend = iend+ len
        read(iounit,valfmt) (rhs(i), i = next, iend)
      endif
!
 12   close(iounit)
      return
!---------end-of-readmt_c64-------------------------------------------------
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
!     some routines extracted from SPARSKIT2 and BLAS.
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
!-----------------------------------------------------------------------
      function distdot(n,x,ix,y,iy)
      integer n, ix, iy
      COEF distdot, x(*), y(*), ddot
      external ddot
      distdot = ddot(n,x,ix,y,iy)
      return
      end
!-----end-of-distdot
!-----------------------------------------------------------------------

!----------------------------------------------------------------------c
!                          S P A R S K I T                             c
!----------------------------------------------------------------------c
!               REORDERING ROUTINES -- LEVEL SET BASED ROUTINES        c
!----------------------------------------------------------------------c
! dblstr   : doubled stripe partitioner
! rdis     : recursive dissection partitioner
! dse2way  : distributed site expansion usuing sites from dblstr
! dse      : distributed site expansion usuing sites from rdis
!------------- utility routines -----------------------------------------
! BFS      : Breadth-First search traversal algorithm
! add_lvst : routine to add a level -- used by BFS
! stripes  : finds the level set structure
! stripes0 : finds a trivial one-way partitioning from level-sets
! perphn   : finds a pseudo-peripheral node and performs a BFS from it.
! mapper4  : routine used by dse and dse2way to do center expansion
! get_domns: routine to find subdomaine from linked lists found by
!            mapper4.
! add_lk   : routine to add entry to linked list -- used by mapper4.
! find_ctr : routine to locate an approximate center of a subgraph.
! rversp   : routine to reverse a given permutation (e.g., for RCMK)
! maskdeg  : integer function to compute the `masked' of a node
!-----------------------------------------------------------------------
      subroutine dblstr(n,ja,ia,ip1,ip2,nfirst,riord,ndom,map,mapptr, &
           mask,levels,iwk)
      implicit none
      integer ndom,ja(*),ia(*),ip1,ip2,nfirst,riord(*),map(*),mapptr(*), &
           mask(*),levels(*),iwk(*),nextdom
!-----------------------------------------------------------------------
!     this routine does a two-way partitioning of a graph using
!     level sets recursively. First a coarse set is found by a
!     simple cuthill-mc Kee type algorithm. Them each of the large
!     domains is further partitioned into subsets using the same
!     technique. The ip1 and ip2 parameters indicate the desired number
!     number of partitions 'in each direction'. So the total number of
!     partitions on return ought to be equal (or close) to ip1*ip2
!----------------------parameters----------------------------------------
! on entry:
!---------
! n      = row dimension of matrix == number of vertices in graph
! ja, ia = pattern of matrix in CSR format (the ja,ia arrays of csr data
!          structure)
! ip1    = integer indicating the number of large partitions ('number of
!          paritions in first direction')
! ip2    = integer indicating the number of smaller partitions, per
!          large partition, ('number of partitions in second direction')
! nfirst = number of nodes in the first level that is input in riord
! riord  = (also an ouput argument). on entry riord contains the labels
!          of the nfirst nodes that constitute the first level.
! on return:
!-----------
! ndom   = total number of partitions found
! map    = list of nodes listed partition by partition from partition 1
!          to paritition ndom.
! mapptr = pointer array for map. All nodes from position
!          k1=mapptr(idom),to position k2=mapptr(idom+1)-1 in map belong
!          to partition idom.
! work arrays:
!-------------
! mask   = array of length n, used to hold the partition number of each
!          node for the first (large) partitioning.
!          mask is also used as a marker of  visited nodes.
! levels = integer array of length .le. n used to hold the pointer
!          arrays for the various level structures obtained from BFS.
!
!-----------------------------------------------------------------------
      integer n, j,idom,kdom,jdom,maskval,k,nlev,init,ndp1,numnod
      maskval = 1
      do j=1, n
         mask(j) = maskval
      enddo
      iwk(1) = 0
      call BFS(n,ja,ia,nfirst,iwk,mask,maskval,riord,levels,nlev)
!
!     init = riord(1)
!     call perphn (ja,ia,mask,maskval,init,nlev,riord,levels)
      call stripes (nlev,riord,levels,ip1,map,mapptr,ndom)
!-----------------------------------------------------------------------
      if (ip2 .eq. 1) return
      ndp1 = ndom+1
!
!     pack info into array iwk
!
      do j = 1, ndom+1
         iwk(j) = ndp1+mapptr(j)
      enddo
      do j=1, mapptr(ndom+1)-1
         iwk(ndp1+j) = map(j)
      enddo
      do idom=1, ndom
         j = iwk(idom)
         numnod = iwk(idom+1) - iwk(idom)
         init = iwk(j)
         do k=j, iwk(idom+1)-1
         enddo
      enddo

      do idom=1, ndom
         do k=mapptr(idom),mapptr(idom+1)-1
            mask(map(k)) = idom
         enddo
      enddo
      nextdom = 1
!
!     jdom = counter for total number of (small) subdomains
!
      jdom = 1
      mapptr(jdom) = 1
!-----------------------------------------------------------------------
      do idom =1, ndom
         maskval = idom
         nfirst = 1
         numnod = iwk(idom+1) - iwk(idom)
         j = iwk(idom)
         init = iwk(j)
         nextdom = mapptr(jdom)
!  note:    old version uses iperm array
         call perphn(numnod,ja,ia,init,mask,maskval, &
              nlev,riord,levels)
!
         call stripes (nlev,riord,levels,ip2,map(nextdom), &
              mapptr(jdom),kdom)
!
         mapptr(jdom) = nextdom
         do j = jdom,jdom+kdom-1
            mapptr(j+1) = nextdom + mapptr(j+1)-1
         enddo
         jdom = jdom + kdom
      enddo
!
      ndom = jdom - 1
      return
      end
!-----------------------------------------------------------------------
      subroutine rdis(n,ja,ia,ndom,map,mapptr,mask,levels,size,iptr)
      implicit none
!-----
      integer n,ja(*),ia(*),ndom,map(*),mapptr(*),mask(*),levels(*), &
           size(ndom+1),iptr(ndom+1)
!-----------------------------------------------------------------------
!     recursive dissection algorithm for partitioning.
!     initial graph is cut in two - then each time, the largest set
!     is cut in two until we reach desired number of domains.
!-----------------------------------------------------------------------
!     input
!     n, ja, ia = graph
!     ndom      = desired number of subgraphs
!     output
!     ------
!     map, mapptr  = pointer array data structure for domains.
!     if k1 = mapptr(i), k2=mapptr(i+1)-1 then
!     map(k1:k2) = points in domain number i
!     work arrays:
!    -------------
!    mask(1:n)    integer
!    levels(1:n)  integer
!    size(1:ndom) integer
!    iptr(1:ndom) integer
!-----------------------------------------------------------------------
      integer idom,maskval,k,nlev,init,nextsiz,wantsiz,lev,ko, &
           maxsiz,j,nextdom
!-----------------------------------------------------------------------
      idom = 1
!-----------------------------------------------------------------------
!     size(i) = size of domnain  i
!     iptr(i)  = index of first element of domain i
!-----------------------------------------------------------------------
      size(idom) = n
      iptr(idom) = 1
      do j=1, n
         mask(j) = 1
      enddo
!
!     domain loop
!
 1    continue
!
!     select domain with largest size
!
      maxsiz = 0
      do j=1, idom
         if (size(j) .gt. maxsiz) then
            maxsiz = size(j)
            nextdom = j
         endif
      enddo
!
!     do a Prphn/ BFS on nextdom
!
      maskval = nextdom
      init = iptr(nextdom)
      call perphn(n,ja,ia,init,mask,maskval,nlev,map,levels)
!
!     determine next subdomain
!
      nextsiz = 0
      wantsiz = maxsiz/2
      idom = idom+1
      lev = nlev
      do while (nextsiz .lt. wantsiz)
         do k = levels(lev), levels(lev+1)-1
            mask(map(k)) = idom
        enddo
        nextsiz = nextsiz + levels(lev+1) - levels(lev)
        lev = lev-1
      enddo
!
      size(nextdom) = size(nextdom) - nextsiz
      size(idom) = nextsiz
!
!     new initial point = last point of previous domain
!
       iptr(idom) = map(levels(nlev+1)-1)
!       iptr(idom) = map(levels(lev)+1)
!      iptr(idom) = 1
!
! alternative
!      lev = 1
!      do while (nextsiz .lt. wantsiz)
!         do k = levels(lev), levels(lev+1)-1
!            mask(map(k)) = idom
!         enddo
!         nextsiz = nextsiz + levels(lev+1) - levels(lev)
!         lev = lev+1
!      enddo
!
!     set size of new domain and adjust previous one
!
!      size(idom) = nextsiz
!      size(nextdom) = size(nextdom) - nextsiz
!      iptr(idom) = iptr(nextdom)
!      iptr(nextdom) = map(levels(lev))

      if (idom .lt. ndom) goto 1
!
!     domains found -- build data structure
!
      mapptr(1) = 1
      do idom=1, ndom
         mapptr(idom+1) = mapptr(idom) + size(idom)
      enddo
      do k=1, n
         idom = mask(k)
         ko = mapptr(idom)
         map(ko) = k
         mapptr(idom) = ko+1
      enddo
!
!     reset pointers
!
      do j = ndom,1,-1
         mapptr(j+1) = mapptr(j)
      enddo
      mapptr(1) = 1
!
      return
      end
!-----------------------------------------------------------------------
      subroutine dse2way(n,ja,ia,ip1,ip2,nfirst,riord,ndom,dom,idom, &
           mask,jwk,link)
!-----------------------------------------------------------------------
!     uses centers obtained from dblstr partition to get new partition
!-----------------------------------------------------------------------
!     input: n, ja, ia   = matrix
!     nfirst = number of first points
!     riord  = riord(1:nfirst) initial points
!     output
!     ndom   = number of domains
!     dom, idom = pointer array structure for domains.
!     mask , jwk, link = work arrays,
!-----------------------------------------------------------------------
      implicit none
      integer n, ja(*), ia(*), ip1, ip2, nfirst, riord(*), dom(*), &
           idom(*), mask(*), jwk(n+1),ndom,link(*)
!
!-----------------------------------------------------------------------
!     local variables
      integer i, mid,nsiz, maskval,init, outer, nouter, k
      call dblstr(n,ja,ia,ip1,ip2,nfirst,riord,ndom,dom,idom,mask, &
           link,jwk)
!
      nouter = 3
!-----------------------------------------------------------------------

      do outer =1, nouter
!
!     set masks
!
      do i=1, ndom
         do k=idom(i),idom(i+1)-1
            mask(dom(k)) = i
         enddo
      enddo
!
!     get centers
!
      do i =1, ndom
         nsiz = idom(i+1) - idom(i)
         init = dom(idom(i))
         maskval = i
!
!         use link for local riord -- jwk for other arrays --
!
         call find_ctr(n,nsiz,ja,ia,init,mask,maskval,link, &
          jwk,mid,jwk(nsiz+1))
         riord(i) = mid
      enddo
!
!     do level-set expansion from centers -- save previous diameter
!
      call mapper4(n,ja,ia,ndom,riord,jwk,mask,link)
      call get_domns2(ndom,riord,link,jwk,dom,idom)
!-----------------------------------------------------------------------
      enddo
      return
      end
!-----------------------------------------------------------------------
      subroutine dse(n,ja,ia,ndom,riord,dom,idom,mask,jwk,link)
      implicit none
      integer n, ja(*), ia(n+1), ndom, riord(n), dom(*), &
           idom(ndom+1),mask(n),jwk(2*ndom),link(n)
!-----------------------------------------------------------------------
!     uses centers produced from rdis to get a new partitioning --
!     see calling sequence in rdis..
!     on entry:
!     n, ja, ia = graph
!     ndom = number of desired subdomains
!     on return
!     dom, idom =
!          pointer array structure for the ndom domains.
!
!-----------------------------------------------------------------------
!     size of array jwk = 2*ndom
!     size array link
!     dom = array of size at least n
!     mask = array of size n.
!     link = same size as riord
!     riord = n list of nodes listed in sequence for all subdomains
!-----------------------------------------------------------------------
!     local variables
      integer i, mid, nsiz, maskval,init, outer, nouter, k
!-----------------------------------------------------------------------
      nouter = 3
!
      call rdis(n,ja,ia,ndom,dom,idom,mask,link,jwk,jwk(ndom+1))
!
!     initial points =
!
      do outer =1, nouter
!
!     set masks
!
      do i=1, ndom
         do k=idom(i),idom(i+1)-1
            mask(dom(k)) = i
         enddo
      enddo
!
!     get centers
!
      do i =1, ndom
         nsiz = idom(i+1) - idom(i)
         init = dom(idom(i))
         maskval = i
!
!         use link for local riord -- jwk for other arrays --
!
         call find_ctr(n,nsiz,ja,ia,init,mask,maskval,link, &
          jwk,mid,jwk(nsiz+1))
         riord(i) = mid
      enddo
!
!     do level-set expansion from centers -- save previous diameter
!
      call mapper4(n,ja,ia,ndom,riord,jwk,mask,link)
      call get_domns2(ndom,riord,link,jwk,dom,idom)
!-----------------------------------------------------------------------
      enddo
      return
      end
!-----------------------------------------------------------------------
      subroutine BFS(n,ja,ia,nfirst,iperm,mask,maskval,riord,levels, &
           nlev)
      implicit none
      integer n,ja(*),ia(*),nfirst,iperm(n),mask(n+1),riord(*), &
           levels(*), &
           nlev,maskval
!-----------------------------------------------------------------------
! finds the level-structure (breadth-first-search or CMK) ordering for a
! given sparse matrix. Uses add_lvst. Allows an set of nodes to be
! the initial level (instead of just one node).
!-------------------------parameters------------------------------------
! on entry:
!---------
!     n      = number of nodes in the graph
!     ja, ia = pattern of matrix in CSR format (the ja,ia arrays of csr data
!     structure)
!     nfirst = number of nodes in the first level that is input in riord
!     iperm  = integer array indicating in which order to  traverse the graph
!     in order to generate all connected components.
!     if iperm(1) .eq. 0 on entry then BFS will traverse the nodes
!     in the  order 1,2,...,n.
!
!     riord  = (also an ouput argument). On entry riord contains the labels
!     of the nfirst nodes that constitute the first level.
!
!     mask   = array used to indicate whether or not a node should be
!     condidered in the graph. see maskval.
!     mask is also used as a marker of  visited nodes.
!
!     maskval= consider node i only when:  mask(i) .eq. maskval
!     maskval must be .gt. 0.
!     thus, to consider all nodes, take mask(1:n) = 1.
!     maskval=1 (for example)
!
!     on return
!     ---------
!     mask   = on return mask is restored to its initial state.
!     riord  = `reverse permutation array'. Contains the labels of the nodes
!     constituting all the levels found, from the first level to
!     the last.
!     levels = pointer array for the level structure. If lev is a level
!     number, and k1=levels(lev),k2=levels(lev+1)-1, then
!     all the nodes of level number lev are:
!     riord(k1),riord(k1+1),...,riord(k2)
!     nlev   = number of levels found
!-----------------------------------------------------------------------
!
      integer j, ii, nod, istart, iend
      logical permut
      permut = (iperm(1) .ne. 0)
!
!     start pointer structure to levels
!
      nlev   = 0
!
!     previous end
!
      istart = 0
      ii = 0
!
!     current end
!
      iend = nfirst
!
!     intialize masks to zero -- except nodes of first level --
!
      do 12 j=1, nfirst
!         print *,'mask',riord(j),j,nfirst
         mask(riord(j)) = 0
 12   continue
!-----------------------------------------------------------------------
 13   continue
!
 1    nlev = nlev+1
      levels(nlev) = istart + 1
      call add_lvst (istart,iend,nlev,riord,ja,ia,mask,maskval)
      if (istart .lt. iend) goto 1
 2    ii = ii+1
      if (ii .le. n) then
         nod = ii
         if (permut) nod = iperm(nod)
         if (mask(nod) .eq. maskval) then
!
!     start a new level
!
            istart = iend
            iend = iend+1
            riord(iend) = nod
            mask(nod) = 0
            goto 1
         else
            goto 2
         endif
      endif
!-----------------------------------------------------------------------
 3    levels(nlev+1) = iend+1
      do j=1, iend
         mask(riord(j)) = maskval
      enddo
!-----------------------------------------------------------------------
      return
      end
!-----------------------------------------------------------------------
      subroutine add_lvst(istart,iend,nlev,riord,ja,ia,mask,maskval)
      integer nlev, nod, riord(*), ja(*), ia(*), mask(*)
!-------------------------------------------------------------
!     adds one level set to the previous sets..
!     span all nodes of previous mask
!-------------------------------------------------------------
      nod = iend
      do 25 ir = istart+1,iend
         i = riord(ir)
         do 24 k=ia(i),ia(i+1)-1
            j = ja(k)
            if (mask(j) .eq. maskval) then
               nod = nod+1
               mask(j) = 0
               riord(nod) = j
            endif
 24      continue
 25   continue
      istart = iend
      iend   = nod
      return
      end
!-----------------------------------------------------------------------
      subroutine stripes (nlev,riord,levels,ip,map,mapptr,ndom)
      implicit none
      integer nlev,riord(*),levels(nlev+1),ip,map(*), &
          mapptr(*), ndom
!-----------------------------------------------------------------------
!    this is a post processor to BFS. stripes uses the output of BFS to
!    find a decomposition of the adjacency graph by stripes. It fills
!    the stripes level by level until a number of nodes .gt. ip is
!    is reached.
!---------------------------parameters-----------------------------------
! on entry:
! --------
! nlev   = number of levels as found by BFS
! riord  = reverse permutation array produced by BFS --
! levels = pointer array for the level structure as computed by BFS. If
!          lev is a level number, and k1=levels(lev),k2=levels(lev+1)-1,
!          then all the nodes of level number lev are:
!                      riord(k1),riord(k1+1),...,riord(k2)
!  ip    = number of desired partitions (subdomains) of about equal size.
!
! on return
! ---------
! ndom     = number of subgraphs (subdomains) found
! map      = node per processor list. The nodes are listed contiguously
!            from proc 1 to nproc = mpx*mpy.
! mapptr   = pointer array for array map. list for proc. i starts at
!            mapptr(i) and ends at mapptr(i+1)-1 in array map.
!-----------------------------------------------------------------------
! local variables.
!
      integer ib,ktr,ilev,k,nsiz,psiz
      ndom = 1
      ib = 1
! to add: if (ip .le. 1) then ...
      nsiz = levels(nlev+1) - levels(1)
      psiz = (nsiz-ib)/max(1,(ip - ndom + 1)) + 1
      mapptr(ndom) = ib
      ktr = 0
      do 10 ilev = 1, nlev
!
!     add all nodes of this level to domain
!
         do 3 k=levels(ilev), levels(ilev+1)-1
            map(ib) = riord(k)
            ib = ib+1
            ktr = ktr + 1
            if (ktr .ge. psiz  .or. k .ge. nsiz) then
               ndom = ndom + 1
               mapptr(ndom) = ib
               psiz = (nsiz-ib)/max(1,(ip - ndom + 1)) + 1
               ktr = 0
            endif
!
 3       continue
 10   continue
      ndom = ndom-1
      return
      end
!-----------------------------------------------------------------------
      subroutine stripes0 (ip,nlev,il,ndom,iptr)
      integer ip, nlev, il(*), ndom, iptr(*)
!-----------------------------------------------------------------------
!     This routine is a simple level-set partitioner. It scans
!     the level-sets as produced by BFS from one to nlev.
!     each time the number of nodes in the accumulated set of
!     levels traversed exceeds the parameter ip, this set defines
!     a new subgraph.
!-------------------------parameter-list---------------------------------
! on entry:
! --------
! ip     = desired number of nodes per subgraph.
! nlev   = number of levels found  as output by BFS
! il     = integer array containing the pointer array for
!          the level data structure as output by BFS.
!          thus il(lev+1) - il(lev) = the number of
!          nodes that constitute the level numbe lev.
! on return
! ---------
! ndom   = number of sungraphs found
! iptr   = pointer array for the sugraph data structure.
!          thus, iptr(idom) points to the first level that
!          consistutes the subgraph number idom, in the
!          level data structure.
!-----------------------------------------------------------------------
      ktr = 0
      iband = 1
      iptr(iband) = 1
!-----------------------------------------------------------------------

      do 10 ilev = 1, nlev
         ktr = ktr + il(ilev+1) - il(ilev)
         if (ktr .gt. ip) then
            iband = iband+1
            iptr(iband) = ilev+1
            ktr = 0
         endif
!
 10   continue
!-----------returning --------------------
      iptr(iband) = nlev + 1
      ndom = iband-1
      return
!-----------------------------------------------------------------------
!-----end-of-stripes0---------------------------------------------------
      end
!-----------------------------------------------------------------------
      integer function maskdeg (ja,ia,nod,mask,maskval)
      implicit none
      integer ja(*),ia(*),nod,mask(*),maskval
!-----------------------------------------------------------------------
      integer deg, k
      deg = 0
      do k =ia(nod),ia(nod+1)-1
         if (mask(ja(k)) .eq. maskval) deg = deg+1
      enddo
      maskdeg = deg
      return
      end
!-----------------------------------------------------------------------
      subroutine perphn(n,ja,ia,init,mask,maskval,nlev,riord,levels)
      implicit none
      integer n,ja(*),ia(*),init,mask(*),maskval, &
           nlev,riord(*),levels(*)
!-----------------------------------------------------------------------
!     finds a peripheral node and does a BFS search from it.
!-----------------------------------------------------------------------
!     see routine  dblstr for description of parameters
! input:
!-------
! ja, ia  = list pointer array for the adjacency graph
! mask    = array used for masking nodes -- see maskval
! maskval = value to be checked against for determing whether or
!           not a node is masked. If mask(k) .ne. maskval then
!           node k is not considered.
! init    = init node in the pseudo-peripheral node algorithm.
!
! output:
!-------
! init    = actual pseudo-peripherial node found.
! nlev    = number of levels in the final BFS traversal.
! riord   =
! levels  =
!-----------------------------------------------------------------------
      integer j,nlevp,deg,nfirst,mindeg,nod,maskdeg
      integer iperm(1)
      nlevp = 0
 1    continue
      riord(1) = init
      nfirst = 1
      iperm(1) = 0
!
      call BFS(n,ja,ia,nfirst,iperm,mask,maskval,riord,levels,nlev)
      if (nlev .gt. nlevp) then
         mindeg = n+1
         do j=levels(nlev),levels(nlev+1)-1
            nod = riord(j)
            deg = maskdeg(ja,ia,nod,mask,maskval)
            if (deg .lt. mindeg) then
               init = nod
               mindeg = deg
            endif
         enddo
         nlevp = nlev
         goto 1
      endif
      return
      end
!-----------------------------------------------------------------------
      subroutine mapper4 (n,ja,ia,ndom,nodes,levst,marker,link)
      implicit none
      integer n,ndom,ja(*),ia(*),marker(n),levst(2*ndom), &
           nodes(*),link(*)
!-----------------------------------------------------------------------
!     finds domains given ndom centers -- by doing a level set expansion
!-----------------------------------------------------------------------
!     on entry:
!     ---------
!     n      = dimension of matrix
!     ja, ia = adajacency list of matrix (CSR format without values) --
!     ndom   = number of subdomains (nr output by coarsen)
!     nodes  = array of size at least n. On input the first ndom entries
!              of nodes should contain the labels of the centers of the
!              ndom domains from which to do the expansions.
!
!     on return
!     ---------
!     link  = linked list array for the ndom domains.
!     nodes = contains the list of nodes of the domain corresponding to
!             link. (nodes(i) and link(i) are related to the same node).
!
!     levst = levst(j) points to beginning of subdomain j in link.
!
!     work arrays:
!     -----------
!     levst : work array of length 2*ndom -- contains the beginning and
!     end of  current level in link.
!     beginning of last level in link for each processor.
!     also ends in levst(ndom+i)
!     marker : work array of length n.
!
!     Notes on implementation:
!     -----------------------
!     for j .le. ndom link(j) is <0  and indicates the end of the
!     linked list. The most recent element added to the linked
!     list is added at the end of the list (traversal=backward)
!     For  j .le. ndom, the value of -link(j) is the size of
!     subdomain j.
!
!-----------------------------------------------------------------------
!     local variables
      integer mindom,j,lkend,nod,nodprev,idom,next,i,kk,ii,ilast,nstuck, &
           isiz, nsize
!
!     initilaize nodes and link arrays
!
      do 10 j=1, n
         marker(j) = 0
 10   continue
!
      do 11 j=1, ndom
         link(j) = -1
         marker(nodes(j)) = j
         levst(j) = j
         levst(ndom+j) = j
 11   continue
!
!     ii = next untouched node for restarting new connected component.
!
      ii = 0
!
      lkend = ndom
      nod   = ndom
      nstuck = 0
!-----------------------------------------------------------------------
 100  continue
      idom = mindom(n,ndom,levst,link)
!-----------------------------------------------------------------------
!     begin level-set loop
!-----------------------------------------------------------------------
 3    nodprev = nod
      ilast = levst(ndom+idom)
      levst(ndom+idom) = lkend
      next = levst(idom)
!
!     linked list traversal loop
!
      isiz = 0
      nsize = link(idom)
 1    i = nodes(next)
      isiz = isiz + 1
!
!     adjacency list traversal loop
!
      do 2 kk=ia(i), ia(i+1)-1
         j = ja(kk)
         if (marker(j) .eq. 0) then
            call add_lk(j,nod,idom,ndom,lkend,levst,link,nodes,marker)
         endif
 2    continue
!
!     if last element of the previous level not reached continue
!
      if (next .gt. ilast) then
         next = link(next)
         if (next .gt. 0) goto 1
      endif
!-----------------------------------------------------------------------
!     end level-set traversal --
!-----------------------------------------------------------------------
      if (nodprev .eq. nod) then
!
!     link(idom) >0 indicates that set is stuck  --
!
         link(idom) = -link(idom)
         nstuck = nstuck+1
      endif
!
      if (nstuck .lt. ndom) goto 100
!
!     reset sizes --
!
      do j=1, ndom
         if (link(j) .gt. 0) link(j) = -link(j)
      enddo
!
      if (nod .eq. n) return
!
!     stuck. add first non assigned point to smallest domain
!
 20   ii = ii+1
      if (ii .le. n) then
         if (marker(ii) .eq. 0) then
            idom = 0
            isiz = n+1
            do 30 kk=ia(ii), ia(ii+1)-1
               i = marker(ja(kk))
               if (i .ne. 0) then
                  nsize = abs(link(i))
                  if (nsize .lt. isiz) then
                     isiz = nsize
                     idom = i
                  endif
               endif
 30            continue
!
!     if no neighboring domain select smallest one
!
               if (idom .eq. 0) idom = mindom(n,ndom,levst,link)
!
!     add ii to sudomain idom at end of linked list
!
            call add_lk(ii,nod,idom,ndom,lkend,levst,link,nodes,marker)
            goto 3
         else
            goto 20
         endif
      endif
      return
      end
!-----------------------------------------------------------------------
      subroutine get_domns2(ndom,nodes,link,levst,riord,iptr)
      implicit none
      integer ndom,nodes(*),link(*),levst(*),riord(*),iptr(*)
!-----------------------------------------------------------------------
!     constructs the subdomains from its linked list data structure
!-----------------------------------------------------------------------
!     input:
!     ndom  = number of subdomains
!     nodes = sequence of nodes are produced by mapper4.
!     link  = link list array as produced by mapper4.
!     on return:
!----------
!     riord = contains the nodes in each subdomain in succession.
!     iptr  = pointer in riord for beginnning of each subdomain.
!     Thus subdomain number i consists of nodes
!     riord(k1),riord(k1)+1,...,riord(k2)
!     where k1 = iptr(i), k2= iptr(i+1)-1
!
!-----------------------------------------------------------------------
!     local variables
      integer nod, j, next, ii
      nod = 1
      iptr(1) = nod
      do 21 j=1, ndom
         next = levst(j)
 22      ii = nodes(next)
         riord(nod) = ii
         nod = nod+1
         next = link(next)
         if (next .gt.  0) goto 22
         iptr(j+1) = nod
 21   continue
!
      return
!-----------------------------------------------------------------------
      end
!-----------------------------------------------------------------------
      function mindom(n, ndom, levst, link)
      implicit none
      integer mindom, n, ndom, levst(2*ndom),link(n)
!-----------------------------------------------------------------------
!     returns  the domain with smallest size
!-----------------------------------------------------------------------
!      locals
!
      integer i, nsize, isiz
!
      isiz = n+1
      do 10 i=1, ndom
         nsize = - link(i)
         if (nsize .lt. 0) goto 10
         if (nsize .lt. isiz) then
            isiz = nsize
            mindom = i
         endif
 10   continue
      return
      end
!-----------------------------------------------------------------------
      subroutine add_lk(new,nod,idom,ndom,lkend,levst,link,nodes,marker)
      implicit none
      integer new,nod,idom,ndom,lkend,levst(*),link(*),nodes(*), &
           marker(*)
!-----------------------------------------------------------------------
!     adds from head --
!
!     adds one entry (new) to linked list and ipdates everything.
!     new  = node to be added
!     nod  = current number of marked nodes
!     idom = domain to which new is to be added
!     ndom = total number of domains
!     lkend= location of end of structure (link and nodes)
!     levst= pointer array for link, nodes
!     link = link array
!     nodes= nodes array --
!     marker = marker array == if marker(k) =0 then node k is not
!              assigned yet.
!-----------------------------------------------------------------------
!      locals
!
      integer ktop
      lkend = lkend + 1
      nodes(lkend) = new
      nod = nod+1
      marker(new) = idom
      ktop = levst(idom)
      link(lkend) = ktop
      link(idom) = link(idom)-1
      levst(idom) = lkend
      return
!-----------------------------------------------------------------------
!-------end-of-add_lk---------------------------------------------------
      end
!-----------------------------------------------------------------------
      subroutine find_ctr(n,nsiz,ja,ia,init,mask,maskval,riord, &
           levels,center,iwk)
      implicit none
      integer n,nsiz,ja(*),ia(*),init,mask(*),maskval,riord(*), &
           levels(*),center,iwk(*)
!-----------------------------------------------------------------------
!     finds a center point of a subgraph --
!-----------------------------------------------------------------------
!     n, ja, ia = graph
!     nsiz = size of current domain.
!     init = initial node in search
!     mask
!     maskval
!-----------------------------------------------------------------------
!     local variables
      integer midlev, nlev,newmask, k, kr, kl, init0, nlev0
      call perphn(n,ja,ia,init,mask,maskval,nlev,riord,levels)
!-----------------------------------------------------------------------
!     midlevel = level which cuts domain into 2 roughly equal-size
!     regions
!
      midlev = 1
      k = 0
 1    continue
      k = k + levels(midlev+1)-levels(midlev)
      if (k*2 .lt. nsiz) then
         midlev = midlev+1
         goto 1
      endif
!-----------------------------------------------------------------------
      newmask = n+maskval
!
!     assign temporary masks to mid-level elements
!
      do k=levels(midlev),levels(midlev+1)-1
         mask(riord(k)) = newmask
      enddo
!
!     find pseudo-periph node for mid-level `line'
!
      kr = 1
      kl = kr + nsiz
      init0 = riord(levels(midlev))
      call perphn(n,ja,ia,init0,mask,newmask,nlev0,iwk(kr),iwk(kl))
!-----------------------------------------------------------------------
!     restore  mask to initial state
!-----------------------------------------------------------------------
      do k=levels(midlev),levels(midlev+1)-1
         mask(riord(k)) = maskval
      enddo
!-----------------------------------------------------------------------
!     define center
!-----------------------------------------------------------------------
      midlev = 1 + (nlev0-1)/2
      k = iwk(kl+midlev-1)
      center = iwk(k)
!-----------------------------------------------------------------------
      return
      end
!-----------------------------------------------------------------------
      subroutine rversp (n, riord)
      integer n, riord(n)
!-----------------------------------------------------------------------
!     this routine does an in-place reversing of the permutation array
!     riord --
!-----------------------------------------------------------------------
      integer j, k
      do 26 j=1,n/2
         k = riord(j)
         riord(j) = riord(n-j+1)
         riord(n-j+1) = k
 26   continue
      return
      end
!-----------------------------------------------------------------------
