/******************************************************************************/
/* Copyright 2008-2009 CNRS, INRIA                                            */
/*                                                                            */
/* File          : testHIPS-Save.c                                            */
/*                                                                            */
/* Description   : This file is part of HIPS : Hierarchical Iterative         */
/*                 Parallel Solver (http://hips.gforge.inria.fr/).            */
/*                                                                            */
/* Authors       : J. GAIDAMOUR and P. HENON                                  */
/******************************************************************************/

/******************************************************************************/
/* License       : CeCILL-C                                                   */
/*                                                                            */
/* This software is governed by the CeCILL-C license under French law and     */
/* abiding by the rules of distribution of free software. You can use,        */
/* modify and/or redistribute the software under the terms of the             */
/* CeCILL-C license as circulated by CEA, CNRS and INRIA at the following     */
/* URL: "http://www.cecill.info".                                             */
/*                                                                            */
/* As a counterpart to the access to the source code and rights to copy,      */
/* modify and redistribute granted by the license, users are provided         */
/* only with a limited warranty and the software's author, the holder of      */
/* the economic rights, and the successive licensors have only limited        */
/* liability.                                                                 */
/*                                                                            */
/* In this respect, the user's attention is drawn to the risks associated     */
/* with loading, using, modifying and/or developing or reproducing the        */
/* software by the user in light of its specific status of free software,     */
/* that may mean that it is complicated to manipulate, and that also          */
/* therefore means that it is reserved for developers and experienced         */
/* professionals having in-depth computer knowledge. Users are therefore      */
/* encouraged to load and test the software's suitability as regards          */
/* their requirements in conditions enabling the security of their            */
/* systems and/or data to be ensured and, more generally, to use and          */
/* operate it in the same conditions as regards security.                     */
/*                                                                            */
/* The fact that you are presently reading this means that you have had       */
/* knowledge of the CeCILL-C license and that you accept its terms.           */
/******************************************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <mpi.h>
#include "math.h"

#include "hips.h"
#include "io.h"

#define BUFLEN 200

int main(int argc, char *argv[])
{  

  /* to read parameters */
  int  sym_pattern, sym_matrix;
  char matrixfile[BUFLEN];
  char rhsfile   [BUFLEN];
  /*  */

  /*  */
  int id, idnbr/* , i */;
  int n/* , ln */;
  INTL nnz;
  INTL *ia; int *ja; /* INTL *lia; int *lja; */
  COEF *a/* , *la */;
  /*   int interior; */
  INTS domsize=-1, nbproc;
  INTS ierr;
  /*  */

  /** Init MPI environment **/
  MPI_Init(&argc, &argv);
  /*   MPI_Comm_rank(MPI_COMM_WORLD, &proc_id); */


  if(argc < 2)
    {
      fprintf(stderr, "Illegal argument : %s <nb proc> <domain size (in number of node)> \n", argv[0]);
      exit(-1);
    }
  
  nbproc = atoi(argv[1]);
  if(argc > 2)
    domsize = atoi(argv[2]);


  /************************************/
  /* Initialize HIPS for one problem  */
  /************************************/
  idnbr = 1; /* total */
  HIPS_Initialize(idnbr);

  id = 0; /** id of the linear system **/


  /******************************************/
  /* Read the parameters from "inputs" file */
  /******************************************/
  ierr = HIPS_ReadOptionsFromFile(id, NULL, &sym_pattern, &sym_matrix, matrixfile, rhsfile);
  HIPS_ExitOnError(ierr);

  /** parameter domsize is an argument of testHIPS.ex **/
  if(argc > 2)
    {
      HIPS_SetOptionINT(id, HIPS_PARTITION_TYPE, 0);
      HIPS_SetOptionINT(id, HIPS_DOMSIZE, domsize);
    }
  else
    {
      HIPS_SetOptionINT(id, HIPS_DOMNBR, nbproc);
    }


  /**********************************/
  /* Read the matrix from file      */
  /**********************************/
  ierr = CSRread(matrixfile, &n, &nnz, &ia, &ja, &a, &sym_matrix);
  HIPS_ExitOnError(ierr);


  HIPS_SetOptionINT (id, HIPS_SYMMETRIC, sym_matrix);

  /** C : numbering starts from 0 **/
  CSR_Fnum2Cnum(ja, ia, n);
  HIPS_SetOptionINT(id, HIPS_FORTRAN_NUMBERING, 0);

  /***************************************************/
  /*                                                 */
  /* ENTER THE MATRIX GRAPH                          */
  /*                                                 */
  /***************************************************/
  HIPS_GraphGlobalCSR(id, n, ia, ja, 0);
  
  
 

  /***************************************/
  /* Dump to disk */
  /***************************************/
  HIPS_SetupSave(id, "./hid");

  HIPS_LocalMatricesSave(id, nbproc, n, ia, ja, a, "./mtx");


  free(ia);
  free(ja);
  free(a);

  /************************************************/
  /* Free HIPS internal structure for problem id  */
  /************************************************/
  HIPS_Clean(id);


  /**********************************/
  /* Free HIPS internal structures  */
  /**********************************/
  HIPS_Finalize();


  /** End MPI **/
  MPI_Finalize();

  fprintf(stdout, "End\n");  
  
  return 0;
}
